	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Convert/Convert 3D to 2D Polys.py;
	# Created by Bill Wood 16/2/21
	
	class Params(metaclass = Marionette.OrderedClass):
	#APPEARANCE
		#Name
		this = Marionette.Node( "Convert 3D to 2D Polys" )
		this.SetDescription( "Converts 3D polygon(s) to 2D planar polygon(s)." )
	
		#Input Ports
		obj = Marionette.PortIn( vs.Handle(0), '3D poly(s)' )
		obj.SetDescription( "an 3D poly object(s) to be converted" )
	
		flipsides = Marionette.PortIn(False, 'bFlipsides' )
		flipsides.SetDescription( "flip sides of 3D poly object(s)" )
	
		#OIP Controls
	
		#Output Ports
		polys = Marionette.PortOut('2D poly(s)')
		polys.SetDescription( "the 2D polygon(s) after the conversion which can be passed to Extrude Path Input" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
	
		def createfrom3Dpoly(srcpolyH,flipsides):
			X = 0
			Y = 1
			Z = 2
	
			def getmatrix(objH,pt1,pt2,cnt):
				v1 = (pt2[X]-pt1[X],pt2[Y]-pt1[Y],pt2[Z]-pt1[Z])
				if cnt >= 3:
					pt3 = vs.GetPolyPt3D(objH,2)
					v2 = (pt3[X]-pt1[X],pt3[Y]-pt1[Y],pt3[Z]-pt1[Z])
				else: #single line polygon
					v2 = vs.Perp(v1)
					if v2 == (0,0,0): # vertical
						v2 = (0,1,0)
	
				v3 = vs.CrossProduct(v1, v2)
				v2 = vs.CrossProduct(v1, v3)
				if not flipsides:
					v2 = (-v2[X],-v2[Y],-v2[Z])
			
				mati = vs.UnitVec(v1)
				matj = vs.UnitVec(v2)
				matk = vs.UnitVec(v3)
			
				return (mati,matj,matk)
			
			def getpolypath(objH,pcnt,v1,v2):
				base = (v2[X]-v1[X],v2[Y]-v1[Y],v2[Z]-v1[Z])
				vs.ClosePoly()
				vs.BeginPoly()
				
				vs.AddPoint(0,0)
				vlen = vs.Norm(base)
				vs.AddPoint(vlen,0)
				
				for i in range(pcnt):
					v3 = vs.GetPolyPt3D(objH,i)
					pt  = (v3[X]-v1[X],v3[Y]-v1[Y],v3[Z]-v1[Z])
					ang = vs.AngBVec(base,pt)
					vlen = vs.Norm(pt)
					v3  = vs.Ang2Vec(ang,vlen)
					vs.AddPoint(v3[X],v3[Y])
					
				vs.EndPoly()
				
				return vs.LNewObj() #getpolypath
	
			# =========== createfrom3Dpoly
			pt1 = vs.GetPolyPt3D(srcpolyH,0)
			pt2 = vs.GetPolyPt3D(srcpolyH,1)
			cnt = vs.GetVertNum(srcpolyH)
	
			mat = getmatrix(srcpolyH,pt1,pt2,cnt,flipsides)
			u = mat[0]
			v = mat[1]
			w = mat[2]
			
			pathH = getpolypath(srcpolyH,cnt,pt1,pt2)
			btest = vs.SetEntityMatrixN(pathH,u,v,w,pt1)
			
			return pathH
			# createfrom3Dpoly}
	
		#inputs
		objH = self.Params.obj.value
		flipsides = self.Params.flipsides.value
	
		plist = []
		objty = vs.GetTypeN(objH)
		if (objty == 11): # group of 3D polys already converted
			poly3D = vs.FInGroup(objH)
			while poly3D != None:
				poly2D = createfrom3Dpoly(poly3D,flipsides)
				plist.append(poly2D)
				poly3D = vs.NextObj(poly3D)
			vs.Marionette_DisposeObj(objH)
	
		else: # single 3D poly
			polyH = createfrom3Dpoly(objH)
			plist = [polyH]
	
		#outputs
		self.Params.polys.value = plist