	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Convert/Convert to 3D Lines.py;
	# Created by Bill Wood 25/4/20
	
	class Params(metaclass = Marionette.OrderedClass):
		#name
		this = Marionette.Node( "Convert to 3D Lines" )
		this.SetDescription( "Converts an object to a group of lines(s) with a single line between linked faces. 3d solid (ie Extrude) and 2D shapes (ie rectangle) can be converted" )
	
		#Input Ports
		obj = Marionette.PortIn( vs.Handle(0), 'hObj' )
		obj.SetDescription( "an object to be converted" )
		orientbetweenfaces = Marionette.PortIn(False, 'bOrient')
		orientbetweenfaces.SetDescription( "Orient line between linked face normals" )
		
		#Output Ports
		lines = Marionette.PortOut('hLines')
		lines.SetDescription( "the 2D lines(s) in 3D space after the conversion which can be passed to Extrude Path Input" )
	
		this.SetLinksObjects()
			
	def RunNode(self):
	
		def getmatrix(objH,pt1,pt2):
			btest,x1,y1,z1 = vs.Centroid3D(objH)
			pt3 = (x1,y1,z1)
	
			v1 = (pt2[0]-pt1[0],pt2[1]-pt1[1],pt2[2]-pt1[2])
			v2 = (pt3[0]-pt1[0],pt3[1]-pt1[1],pt3[2]-pt1[2])
			v3 = vs.CrossProduct(v1, v2)
			v2 = vs.CrossProduct(v1, v3)
			v2 = (-v2[0],-v2[1],-v2[2])
	
			mati = vs.UnitVec(v1)
			matj = vs.UnitVec(v2)
			matk = vs.UnitVec(v3)
	
			return(mati,matj,matk,(pt3[0],pt3[1],pt3[2]))
			# getmatrix
	
		def addvertex(vert,fcnt):
			nonlocal verts
			nonlocal vertfaces
	
			def vertinverts(refvert,verts):
				vref = []
				for i in range(len(verts)):
					v1 = (refvert[0],refvert[1],refvert[2])
					v2 = (verts[i][0],verts[i][1],verts[i][2])
					if vs.EqPt3D(v1,v2,0.00001): # used for rounding
						vref = i
						break
				return vref
	
			vref = vertinverts(vert,verts)
			if vref != []:
				if fcnt not in vertfaces[vref]:
					vertfaces[vref].append(fcnt)
			else:
				verts.append(vert) # add to vertex list
				fref = [fcnt]
				vertfaces.append(fref)
				vref = len(verts)-1 # first in vertex list
	
			return vref
			# addvertex
	
		def setedgevec(edge):
			nonlocal verts
			nonlocal edgevecs
	
			pt1 = verts[edge[0]]
			pt2 = verts[edge[1]]
			
			vec = (pt2[0]-pt1[0],pt2[1]-pt1[1],pt2[2]-pt1[2])
			midpos = (pt1[0]+vec[0]/2,pt1[1]+vec[1]/2,pt1[2]+vec[2]/2)
			edgevec = vs.UnitVec(vec)
	
			# store edge mid position and edge delta X axis direction
			edgevecs.append((midpos,edgevec))
	
			return
			# setedgevec
				
		def addedge(edge,fcnt):
			nonlocal edges
			nonlocal edgefaces
	
			if edge[0] > edge[1]: # order low to high
				edge = (edge[1],edge[0])
	
			if edge in edges:
				edgeref = edges.index(edge)
				if fcnt not in edgefaces[edgeref]:
					edgefaces[edgeref].append(fcnt)
				else:
					fdef = [fcnt]
					edgefaces[edgeref] = fdef # new entry
			else:
				edges.append(edge) # add to edge list
				setedgevec(edge)
	
				fdef = [fcnt]
				edgefaces.append(fdef)
	
			return
			# addedge
	
		def getfacedef(faceH,fcnt):
			nonlocal faces
	
			pt1 = vs.GetPolyPt3D(faceH,0)
			pt2 = vs.GetPolyPt3D(faceH,1)
			xaxis,yaxis,zaxis,centpos = getmatrix(faceH,pt1,pt2)
	
			vlocfrom = addvertex(pt1,fcnt)
			startfrom = vlocfrom
			vlocto   = addvertex(pt2,fcnt)
			
			addedge((vlocfrom,vlocto),fcnt)
			faces.append((xaxis,yaxis,zaxis,centpos))
			vlocfrom = vlocto
	
			vcnt = vs.GetVertNum(faceH)
			for i in range(2,vcnt):
				vert = vs.GetPolyPt3D(faceH, i)
				vlocto   = addvertex(vert,fcnt)
				addedge((vlocfrom,vlocto),fcnt)
				vlocfrom = vlocto
	
			addedge((vlocfrom,startfrom),fcnt) # close polygon
	
			return
			#getfacedef
	
		def drawline(startpos,linelen,lenvec,uvec,classname):
			vs.SetWorkingPlaneN(startpos[0],startpos[1],startpos[2], uvec[0], uvec[1], uvec[2], lenvec[0], lenvec[1], lenvec[2])
			vs.MoveTo(0,0)
			vs.LineTo(linelen,0)
			lineH = vs.LNewObj()
			vs.SetWorkingPlane(x,y,z,xrot,yrot,zrot)
	
			return lineH
	
		def addvec(dir,axis):
			return (dir[0] + axis[0],dir[1] + axis[1],dir[2] + axis[2])
			
		def getpolydefs(grpH):
			polys = []
			objH = vs.FInGroup(grpH)
			while objH != vs.Handle(0):
				tempH = vs.NextObj(objH)
				polys.append(objH)
				objH = tempH
			return polys
	
		# RunNode
		objH = self.Params.obj.value
		orientbetweenfaces = self.Params.orientbetweenfaces.value
		classname = vs.GetClass(objH)
	
		grpPoly = vs.ConvertTo3DPolys(objH) # class of original object transfered to new 3D polys
		if grpPoly:
	
			faces = [] 		# face entry 	= (xaxis,yaxis,zaxis,centpos)
			edges = [] 		# edge entry 	= (fromvertno,tovertno)
			edgevecs  = [] 	# edge entry 	= (midpos,edgevec)
			edgefaces = [] 	# edge entry 	= (faceno1,faceno2,..faceN)
			verts = [] 		# vertex entry 	= (vertexpos)
			vertfaces = []	# vertex entry	= (faceno1,faceno2,..faceN)
			
			lines = []		# resulting lines
	
			polys = getpolydefs(grpPoly)
			polynum = len(polys)
			for i in range(polynum):
				getfacedef(polys[i],i)
			
			(x,y,z,xrot,yrot,zrot) = vs.GetWorkingPlane() # store current WP
			for i in range(len(edges)):
				edge = edges[i]
				startpos 	= verts[edge[0]]
				endpos 		= verts[edge[1]]
				linelen 	= vs.Distance3D(startpos[0],startpos[1],startpos[2],endpos[0],endpos[1],endpos[2])
				xdir = edgevecs[i][1]
	
				# get combined zdir of all linked faces
				if orientbetweenfaces:
					zdir = (0,0,0)
					edgedef = edgefaces[i]
					for faceno in edgedef:
						xaxis,yaxis,zaxis,centpos = faces[faceno]
						zdir = addvec(zdir,zaxis)	
					zdir = vs.UnitVec(zdir)
					ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
					lineH = drawline(startpos,linelen,xdir,zdir,classname)
					lines.append(lineH)
	
				else: # draw line for each linked face
					edgedef = edgefaces[i]
					face1 = edgedef[0]
					face2 = edgedef[1]
					xaxis1,yaxis1,zaxis1,centpos = faces[face1]
					xaxis2,yaxis2,zaxis2,centpos = faces[face2]
					
					zdir = vs.UnitVec(zaxis1)
					ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
					lineH = drawline(startpos,linelen,xdir,zdir,classname)
					lines.append(lineH)
					
					if not vs.EqPt3D((zaxis1[0],zaxis1[1],zaxis1[2]),(zaxis2[0],zaxis2[1],zaxis2[2]),0.001): # inlinefaces so omit second line
						zdir = vs.UnitVec(zaxis2)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						lineH = drawline(startpos,linelen,xdir,zdir,classname)
						lines.append(lineH)
		else:
			lines = vs.Handle(0)
	
		vs.Marionette_DisposeObj(grpPoly)
	
		self.Params.lines.value = lines