	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Convert/Convert to Loci.py;
	# Created by Bill Wood 24/2/21
	
	class Params(metaclass = Marionette.OrderedClass):
	
		def getboolval(bvalue):
			if isinstance(bvalue,list):
				return bvalue[0]
			else:
				return bvalue
	
		def getnumbval(value,defvalue):
			if isinstance(value,(list,tuple)):
				value = value[0]
			if isinstance(value,(int,float)):
				return value
			else:
				return defvalue
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "Convert to Loci" )
		# this.SetListAbsorb()			   mark this node as list absorbing
		this.SetDescription( "Converts an object to a group of 2D/3D loci." )
	
		#Input Ports
		obj = Marionette.PortIn( vs.Handle(0), 'hObj' )
		obj.SetDescription( "an object to be converted" )
		
		# Note must set atvertices,atmidedges and or atfacecentres
		atvertices = Marionette.PortIn(False, 'bVertices')
		atvertices.SetDescription( "Place objects at poly vertices" )
	
		atmidedges = Marionette.PortIn(False, 'bMidEdges')
		atmidedges.SetDescription( "Place objects at poly mid edges" )
		
		atfacecentres = Marionette.PortIn(False, 'bFaceCentres')
		atfacecentres.SetDescription( "Place objects at poly face centres" )
	
		zoffset = Marionette.PortIn(0, 'nZoffset')
		zoffset.SetDescription( "Z offset of placed objects" )
	
		vertical = Marionette.PortIn(False, 'bVertical')
		vertical.SetDescription( "Placed objects always vertical" )
	
		flipdirection = Marionette.PortIn(False, 'bFlipDir')
		flipdirection.SetDescription( "Placed objects flipped about XY layer plane" )
	
		#Output Ports
		loci = Marionette.PortOut()
		loci.SetDescription( "the 2D/3D loci in 3D space after the conversion" )
	
	
	#BEHAVIOR
		# this.SetLinksObjects()
	
	def RunNode(self):
	
		def getPolydefs(grpPoly):	
			faces = []
			edges = []
			edgevecs  = []
			edgefaces = []
			verts = []
			vertfaces = []
	
			def getmatrix(objH,pt1,pt2):
				# pt3,zval = vs.Get3DCntr(objH)
				btest,x1,y1,z1 = vs.Centroid3D(objH)
				pt3 = (x1,y1,z1)
		
				v1 = (pt2[0]-pt1[0],pt2[1]-pt1[1],pt2[2]-pt1[2])
				v2 = (pt3[0]-pt1[0],pt3[1]-pt1[1],pt3[2]-pt1[2])
				# v2 = (pt3[0]-pt1[0],pt3[1]-pt1[1],zval-pt1[2])
				v3 = vs.CrossProduct(v1, v2)
				v2 = vs.CrossProduct(v1, v3)
				v2 = (-v2[0],-v2[1],-v2[2])
		
				mati = vs.UnitVec(v1)
				matj = vs.UnitVec(v2)
				matk = vs.UnitVec(v3)
		
				return(mati,matj,matk,(pt3[0],pt3[1],pt3[2]))
			
			def addvertex(vert,fcnt):
				nonlocal verts
				nonlocal vertfaces
		
				if vert in verts:
					vref = verts.index(vert)
					if fcnt not in vertfaces[vref]:
						vertfaces[vref].append(fcnt)
					return vref # return entry in list
				else:
					verts.append(vert) # add to vertex list
					fref = [fcnt]
					vertfaces.append(fref)
					return len(verts)-1
		
			def addedge(edge,fcnt):
				nonlocal edges
				nonlocal edgefaces
	
				def setedgevec(edge):
					nonlocal verts
					nonlocal edgevecs
					
					pt1 = verts[edge[0]]
					pt2 = verts[edge[1]]
					
					vec = (pt2[0]-pt1[0],pt2[1]-pt1[1],pt2[2]-pt1[2])
					midpos = (pt1[0]+vec[0]/2,pt1[1]+vec[1]/2,pt1[2]+vec[2]/2)
					
					edgevec = vs.UnitVec(vec)
					# store edge mid position and edge delta X axis direction
					edgevecs.append((midpos,edgevec))
					
					#setedgevec
	
				if edge[0] > edge[1]: # order low to high
					edge = (edge[1],edge[0])
		
				if edge in edges:
					edgeref = edges.index(edge)
					if fcnt not in edgefaces[edgeref]:
						edgefaces[edgeref].append(fcnt)
					else:
						fdef = [fcnt]
						edgefaces[edgeref] = fdef # new entry
				else:
					edges.append(edge) # add to edge list
					setedgevec(edge)
		
					fdef = [fcnt]
					edgefaces.append(fdef)
				return
				
				# addedge
	
			def getfacedef(faceH,fcnt):
				nonlocal faces
	
				pt1 = vs.GetPolyPt3D(faceH,0)
				pt2 = vs.GetPolyPt3D(faceH,1)
				xaxis,yaxis,zaxis,centpos = getmatrix(faceH,pt1,pt2)
	
				vlocfrom = addvertex(pt1,fcnt)
				
				startfrom = vlocfrom
				vlocto   = addvertex(pt2,fcnt)
				
				edgeno   = addedge((vlocfrom,vlocto),fcnt)
				faces.append((xaxis,yaxis,zaxis,centpos))
				vlocfrom = vlocto
	
				vcnt = vs.GetVertNum(faceH)
				for i in range(2,vcnt):
					vert = vs.GetPolyPt3D(faceH, i)
					vlocto   = addvertex(vert,fcnt)
					addedge((vlocfrom,vlocto),fcnt)
					vlocfrom = vlocto
		
				addedge((vlocfrom,startfrom),fcnt) # close polygon	
			
				# getfacedef
	
			# getPolydefs
			if vs.GetTypeN(grpPoly) == 11:
				poly3D = vs.FInGroup(grpPoly)
				i = 0
				while poly3D != None:
					getfacedef(poly3D,i)
					poly3D = vs.NextObj(poly3D)
					i += 1
	
				vs.Marionette_DisposeObj(grpPoly)
			else:
				getfacedef(grpPoly,i) # single 3D polygon
			
			polydef = {'faces':faces,'vertices':verts,'edges':edges,'edgevecs':edgevecs,'edgefaces':edgefaces,'vertfaces':vertfaces}
			
			return polydef
	
			# getPolydefs
	
		def placeloci(zoffset,atvertices,atmidedges,atfacecentres,vertical,flipdirection,polydefs):
			faces		= polydefs['faces']
			vertices	= polydefs['vertices']
			edges		= polydefs['edges']
			edgevecs	= polydefs['edgevecs']
			edgefaces	= polydefs['edgefaces']
			vertfaces	= polydefs['vertfaces']
	
			def createlocus(locdef,zoffset):
				nonlocal lociobjs
				
				axisX,axisY,axisZ,origin = locdef
				if zoffset != 0:
					origin = (origin[0]+axisZ[0]*zoffset,origin[1]+axisZ[1]*zoffset,origin[2]+axisZ[2]*zoffset)
	
				vs.Locus((0,0))
				locusH = vs.LNewObj()
				if vs.SetEntityMatrixN(locusH, axisX, axisY, axisZ, origin):
					lociobjs.append(locusH)
				else:
					vs.DelObject(locusH)
				return
			
			def addvec(dir,axis):
				return (dir[0] + axis[0],dir[1] + axis[1],dir[2] + axis[2])
			
			def flipvec(vec):
				return (-vec[0],-vec[1],-vec[2])
	
			# ===== placeloci
			if atfacecentres:
				for face in faces:
					if vertical:
						xdir = (1,0,0)
						ydir = (0,1,0)
						zdir = (0,0,-1) if flipdirection else (0,0,1)
						origin = face[3]
					else:
						xdir,ydir,zdir,origin = face
	
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
					createlocus((xdir,ydir,zdir,origin),zoffset)
		
			if atvertices:
				if vertical:
					# reset X, Y and Z axes if used
					xdir = (1,0,0); ydir = (0,1,0)
					zdir = (0,0,-1) if flipdirection else (0,0,1)
	
				for vno in range(len(vertices)):
					vertexpos = vertices[vno]
	
					if not vertical:
						xdir = (0,0,0)
						ydir = (0,0,0)
						zdir = (0,0,0)
	
						facenos = vertfaces[vno]
						# vs.AlrtDialog(str(vertfaces))
						for j in range(len(facenos)):
							faceno = facenos[j]
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = addvec(zdir,zaxis)
	
						zdir = vs.UnitVec(zdir)
						
						if vs.EqPt3D((zdir[0],zdir[1],abs(zdir[2])),(0,0,1),0.001): # zdir is vertical
							xdir = (1,0,0) if zdir[2] > 0 else (-1,0,0)
						else:
							xdir = vs.Perp(zdir)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
					
					createlocus((xdir, ydir, zdir, vertexpos),zoffset)
				
			if atmidedges:
				for i in range(len(edgefaces)):
					edgevec = edgevecs[i]
					midpos = edgevec[0]
					
					if vertical:
						# reset X, Y and Z axes
						xdir = (1,0,0); ydir = (0,1,0)
						zdir = (0,0,-1) if flipdirection else (0,0,1)
					else:
						facedef = edgefaces[i]
						xdir	= edgevec[1]
						zdir = (0,0,0)
						for faceno in facedef:
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = addvec(zdir,zaxis)
	
						zdir = vs.UnitVec(zdir)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
	
					createlocus((xdir, ydir, zdir, midpos),zoffset)
	
			return
			# ===== placeloci
	
		# ===== RunNode
		status = 0
		lociobjs = []
	
		# Node controls
		obj = self.Params.obj.value
	
		# UI controls
		atvertices = self.Params.getboolval(self.Params.atvertices.value)
		atmidedges = self.Params.getboolval(self.Params.atmidedges.value)
		atfacecentres = self.Params.getboolval(self.Params.atfacecentres.value)
		atset = (atvertices or atmidedges or atfacecentres)
		
		grpPoly = vs.ConvertTo3DPolys(obj)
		if (grpPoly and atset):
			polydefs = getPolydefs(grpPoly)
	
			zoffset			= self.Params.getnumbval(self.Params.zoffset.value,0)
			vertical	  	= self.Params.getboolval(self.Params.vertical.value)
			flipdirection	= self.Params.getboolval(self.Params.flipdirection.value)
	
			placeloci(zoffset,atvertices,atmidedges,atfacecentres,vertical,flipdirection,polydefs)
	
		#outputs
		self.Params.loci.value = lociobjs
	
	