	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Convert/Offset path.py;
	# Created by Bill Wood 16/2/21
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if (type(inputObject) == vs.Handle): # single object
				objlistH = [inputObject]
			else:
				objlistH = inputObject # list of objects??
			
			newobjlistH = []
			if isinstance(objlistH,list):
				for objH in objlistH:
					objtype = vs.GetTypeN(objH)
					if objtype in [2,5,21]: # 2D polygon polyline objects
						tempH  = vs.CreateDuplicateObject(objH, None) #vs.Handle(0)
						newobjlistH.append(tempH)
	
					elif objtype in [3,4,6,13]: # other 2D objects
						newobjlistH.append(vs.MakePolyline(objH))
	
					elif (objtype == 11): # group at top level
						grpH  = vs.CreateDuplicateObject(objH, None)
						grpobjH = vs.FInGroup(grpH)
						while grpobjH != vs.Handle(0): #scan top level group only
							objtype = vs.GetTypeN(grpobjH)
							nextobjH = vs.NextObj(grpobjH)
							if objtype in [2,5,21]: # 2D path objects
								newobjlistH.append(grpobjH)
							elif objtype in [3,4,6,13]:
								grpobjH = vs.ConvertToPolyline(grpobjH)
								newobjlistH.append(grpobjH)
							else:
								vs.DelObject(grpobjH) # invalid object
							grpobjH = nextobjH
						vs.HUngroup(grpH) 
						
				if len(newobjlistH) == 0:
					return 0 # invalid objects in objectlist
				else:
					return newobjlistH
			else:
				return 0 # not a list
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "Offset Path" )
		this.SetListAbsorb()			  # mark this node as list absorbing
		this.SetDescription( 'Offset path polygons' )
	
		#Input Ports
		polyH = Marionette.PortIn( vs.Handle(0), 'hPoly(s)' )
		polyH.SetDescription( "The input poly(s) seperate or in single level group" )
		offset = Marionette.PortIn( 0, 'nOffset' )
		offset.SetDescription( "The offset value" )
		fillet = Marionette.PortIn( 0, 'nFillet' )
		fillet.SetDescription( "The fillet +ive value" )
	
		#OIP Controls
		hidepath = Marionette.OIPControl('Hide source path', Marionette.WidgetType.Bool, True)
		hidepath.SetDescription( "Hide source path edges in resultant poly(s)" )
	
		#Output Ports
		poly = Marionette.PortOut('hPath(s)')
		poly.SetDescription( "The resultant offset poly(s)" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
		X = 0
		Y = 1
		Z = 2
	
		def getlinexaxis(lineH,offH,rID):
			pt1 = vs.GetSegPt1(lineH)
			pt2 = vs.GetSegPt2(lineH)
			pt3 = vs.GetSegPt1(offH)
	
			btest,loc = vs.PlanarPtTo3DModelPt(rID,pt1)
			btest,pt3D = vs.PlanarPtTo3DModelPt(rID,pt2)
			vX = (pt3D[X]-loc[X],pt3D[Y]-loc[Y],pt3D[Z]-loc[Z])
	
			btest,pt3D = vs.PlanarPtTo3DModelPt(rID,pt3)
			vY = (pt3D[X]-loc[X],pt3D[Y]-loc[Y],pt3D[Z]-loc[Z])
			
			vZ = vs.CrossProduct(vX,vY)
			
			return  (vX,vZ,loc)
	
		def setfillet(poly,start,end,fillet):
			for i in range(start,end):
				pt1,vType,arcRad = vs.GetPolylineVertex(poly, i)
				if vType == 0:
					vs.SetPolylineVertex(poly,i,pt1,3,fillet,True) 
					vs.SetVertexVisibility(poly,i,True)
	
		def setsourcepathvis(poly,vnum):
			for i in range(vnum):
				vs.SetVertexVisibility(poly,i,False)
			vs.SetVertexVisibility(poly,vnum-1,True)
			
		def processpath(srcH,offset,fillet):
			filletset = (fillet > 0)
	
			objty = vs.GetTypeN(srcH)
			lineset = (objty == 2)
			offH   = vs.OffsetPolyN(srcH, offset, False)
	
			if lineset: #Line
				centerPt, normal, uVec = vs.GetWorkingPlaneN() # store WP
				refID = vs.GetPlanarRef(srcH)
				vX,vY,loc = getlinexaxis(srcH,offH,refID)
				vs.SetWorkingPlaneN(loc, vY, vX)
	
				offset = abs(offset)
				linlen = vs.HLength(srcH)
				startpt1 = (0,0)
				endpt1 = (linlen,0)
				startpt2 = (0,offset)
				endpt2 = (linlen,offset)
	
				if hidepath:
					vs.OpenPoly()
				else:
					vs.ClosePoly()
	
				vs.BeginPoly()
	
				vs.MoveTo((startpt1))
				if filletset:
					vs.ArcTo(startpt2,fillet)
					vs.ArcTo(endpt2,fillet)
				else:
					vs.LineTo(startpt2)
					vs.LineTo(endpt2)
				vs.LineTo(endpt1)
	
				vs.EndPoly()
				poly = vs.LNewObj()
	
				vs.DelObject(srcH)
				vs.SetWorkingPlaneN(centerPt, normal, uVec) # restore WP
			
			else: # Polys from getinputObject
				if vs.IsPolyClosed(srcH): #Closed poly
					vnum = 0
					cnt = vs.GetVertNum(offH)
					if (offset > 0):
						poly = vs.ClipSurfaceN(offH,srcH)
						btest,fillH = vs.GetHole(poly,1)
						if hidepath:
							setsourcepathvis(fillH,cnt+1)
						fillH = poly
						vs.DelObject(srcH)
					else:
						vs.SetFPat(srcH,1) # make sure has solid fill white
						poly = vs.ClipSurfaceN(srcH,offH)
						btest,fillH = vs.GetHole(poly,1)
						if hidepath:
							setsourcepathvis(poly,cnt+1)
	
					if filletset:
						setfillet(fillH,vnum+1,cnt+1,fillet)
					
				else: # Open poly
					refID = vs.GetPlanarRef(srcH)
					CentPt, NormVec, UVec = vs.GetWorkingPlaneMat(refID)
					vnum = vs.GetVertNum(srcH)
					
					vs.ClosePoly()
					vs.BeginPoly()
					for i in range(vnum):
						pt1,vtype,arcrad = vs.GetPolylineVertex(srcH,i+1)
						vs.Add2DVertex(pt1,vtype,arcrad)
	
					for i in range(vs.GetVertNum(offH),0,-1):
						pt1,vtype,arcrad = vs.GetPolylineVertex(offH,i)
						if (filletset and (vtype == 0)):
							vtype = 3
							arcrad = fillet
						vs.Add2DVertex(pt1,vtype,arcrad)
					vs.EndPoly()
	
					poly = vs.LNewObj()
					if hidepath:
						setsourcepathvis(poly,vnum)
	
					VVec = vs.CrossProduct(UVec,NormVec)
					btest = vs.SetEntityMatrixN(poly,NormVec,VVec,UVec,CentPt)
	
					fillH = poly
					vs.DelObject(srcH)
	
			vs.DelObject(offH)
			vs.SetDSelect(poly)
			
			return poly
	
		#inputs
		pathlist = []
		pathsH = self.Params.getinputObject(self.Params.polyH.value)
		offset = self.Params.offset.value[0]
		hidepath = self.Params.hidepath.value
	
		if offset != 0:
			fillet = abs(self.Params.fillet.value[0])
			for pathH in pathsH:
				path = processpath(pathH,offset,fillet)
				if path != vs.Handle(0):
					pathlist.append(path)
	
		#outputs
		self.Params.poly.value = pathlist