	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Convert/Update Symbol.py;
	# Created by Bill Wood 5/4/21
	
	class Params(metaclass = Marionette.OrderedClass):
	#APPEARANCE
		#Name
		this = Marionette.Node( "Update Symbol" )
		this.SetDescription( 'This node updates objects in an existing symbol' )
		
		#Input Ports
		objs = Marionette.PortInLists( vs.Handle(0), 'hObj(s)' )
		objs.SetDescription('The objects to add to the symbol') 
	
		inSymname = Marionette.PortIn('', 'sSymname' )
		inSymname.SetDescription('Name of existing symbol')
	
		originobj = Marionette.PortIn( vs.Handle(0), 'hOrigin' )
		originobj.SetDescription('Object defining symbol 3D origin, if unset use bounding box centre') 
		
		#OIP Input Controls
		useorigobj = Marionette.OIPControl('Use Origin Object', Marionette.WidgetType.Bool, False)
		useorigobj.SetDescription( "Use origin object position/rotation for symbol origin" )
		
		BBoxlabel = Marionette.OIPControl( 'Bounding Box Origin', Marionette.WidgetType.Separator, '-')
	
		BBoxZ = ['Upper','Middle','Lower']
		BBoriginZ = Marionette.OIPControl('Point Z', Marionette.WidgetType.Popup, 2, BBoxZ )
		BBoriginZ.SetDescription( "List of bounding box origin options" )
	
		BBoxXY = ['Top Left','Top Centre','Top Right','Centre Left','Centre Centre','Centre Right','Bottom Left','Bottom Centre','Bottom Right']
		BBoriginXY = Marionette.OIPControl('Point XY', Marionette.WidgetType.Popup, 6, BBoxXY )
		BBoriginXY.SetDescription( "List of bounding box origin options" )
		
		#Output Ports
		symbol = Marionette.PortOut('hSymbol')
		symbol.SetDescription('Existing symbol containing new objects')
		
	#BEHAVIOR
	
	def RunNode(self):
		kLINE = 2
		kPOLY = 5
		kARC  = 6
		k3DLOCUS  = 9
		kSYMBOL   = 15
		k2DLOCUS  = 17
		kPOLYLINE = 21
		k3DPOLY   = 25
	
		def chksymexists(symH):
			symexists = False
			if symH != vs.Handle(0):
				symexists = (vs.GetTypeN(symH) == 16) # symbol definition
			return symexists
		
		def chkorigobj(objty):
			return objty in [kLINE,kPOLY,kARC,k3DLOCUS,kSYMBOL,k2DLOCUS,kPOLYLINE,k3DPOLY]
		
		def getoriginobj(origobj,objty):
			def getwdir(x1,y1,x2,y2):
				v1  = (x2-x1,y2-y1)
				wdir = vs.Vec2Ang(v1)
				return wdir
	
			planarset,refid = vs.IsPlanarObj(origobj)
			if planarset:
				btest,pt1,udir,vdir,wdir  = vs.GetEntityMatrix(origobj)
				refID = vs.GetPlanarRef(origobj)
				if objty == kLINE:
					x1,y1 = vs.GetSegPt1(origobj)
					x2,y2 = vs.GetSegPt2(origobj)
					btest,pt1 = vs.PlanarPtTo3DModelPt(refID,(x1,y1))
					wdir = getwdir(x1,y1,x2,y2)
	
				elif objty in [kPOLY,kPOLYLINE]:
					x1,y1 = vs.GetPolyPt(origobj,1)
					x2,y2 = vs.GetPolyPt(origobj,2)
					btest,pt1 = vs.PlanarPtTo3DModelPt(refID,(x1,y1))
					wdir = getwdir(x1,y1,x2,y2)
	
				elif objty == k2DLOCUS:
					btest,pt1 = vs.PlanarPtTo3DModelPt(refID,vs.GetLocPt(origobj))
				#elif objty == kARC: NOT INCLUDED
	
			else:
				best, udir,vdir,wdir, isMirr = vs.Get3DOrientation(origobj)
				if objty == k3DLOCUS:
					pt1 = vs.GetLocus3D(origobj)
					udir = vdir = wdir = 0
				elif objty == kSYMBOL:
					pt1 = vs.GetSymLoc3D(origobj)
				elif objty == k3DPOLY:
					x1,y1,z1 = vs.GetPolyPt3D(origobj,0)
					x2,y2,z2 = vs.GetPolyPt3D(origobj,1)
					pt1 = (x1,y1,z1)
	
			return pt1,(udir,vdir,wdir)
		
		def getoriginbbox(objs,BBoxXY,BBoxZ):
			kTL = 0
			kTC = 1
			kTR = 2
			kCL = 3
			kCC = 4
			kCR = 5
			kBL = 6
			kBC = 7
			kBR = 8
	
			X = 0
			Y = 1
	
			kT = 0
			kC = 1
			kB = 2
	
			if (len(objs) > 1):
				vs.BeginGroup()
				for obj in objs:
					objH = vs.CreateDuplicateObject(obj,None)
					grpH = vs.ConvertTo3DPolys(objH)
					vs.HUngroup(grpH)
				vs.EndGroup()
				grpH = vs.LNewObj()
			else:
				objH = vs.CreateDuplicateObject(objs[0],None)
				grpH = vs.ConvertTo3DPolys(objH)
	
			meshH = vs.GroupToMesh(grpH)
	
			pt1,cz1 = vs.Get3DCntr(meshH)
			cx1 = pt1[X]
			cy1 = pt1[Y]
			hgt, wid, dep = vs.Get3DInfo(meshH)
	
			vs.DelObject(objH)
			vs.DelObject(meshH)
	
			wid /= 2
			hgt /= 2
			dep /= 2
	
			#pTL, pBR
			Lx = cx1-wid
			Cx = cx1
			Rx = cx1+wid
			Ty = cy1+hgt
			Cy = cy1
			By = cy1-hgt
	
			xyloc = [0,0,0,0,0,0,0,0,0]
			xyloc[kTL] = (Lx,Ty)
			xyloc[kTC] = (Cx,Ty)
			xyloc[kTR] = (Rx,Ty)
			
			xyloc[kCL] = (Lx,Cy)
			xyloc[kCC] = (Cx,Cy)
			xyloc[kCR] = (Rx,Cy)
			
			xyloc[kBL] = (Lx,By)
			xyloc[kBC] = (Cx,By)
			xyloc[kBR] = (Rx,By)
	
			#pZ
			zloc = [0,0,0]
			zloc[kT] = cz1+dep
			zloc[kC] = cz1
			zloc[kB] = cz1-dep
	
			pt1 = (xyloc[BBoxXY][X],xyloc[BBoxXY][Y],zloc[BBoxZ])
			return pt1
		
		def getsymobjs(symH):
			symobjs = []
			obj = vs.FInSymDef(symH)
			while obj != vs.Handle(0):
				symobjs.append(obj)
				obj = vs.NextObj(obj)
			return symobjs
	
		def setobjrotpos(objH,objdir,pt1):
			if objdir:
				vs.Set3DRot(objH,objdir,pt1)
			vs.Move3DObj(objH,-pt1[0],-pt1[1],-pt1[2])
	
		#inputs
		objs = self.Params.objs.value
		useorigobj = self.Params.useorigobj.value
		BBoriginZ  = self.Params.BBoriginZ.value
		BBoriginXY = self.Params.BBoriginXY.value
		originobj  = self.Params.originobj.value
		symname = self.Params.inSymname.value
		symH = vs.GetObject(symname)
		symexists = chksymexists(symH)
	
		if objs and symexists:
			objty = vs.GetTypeN(originobj)
			objdir = None
			if useorigobj and chkorigobj(objty):
				pt1,objdir =  getoriginobj(originobj,objty)
			else:
				pt1 =  getoriginbbox(objs,BBoriginXY,BBoriginZ)
	
			symobjs = getsymobjs(symH)
			for obj in objs:
				objH = vs.CreateDuplicateObject(obj, symH)
				setobjrotpos(objH,objdir,pt1)
	
			for obj in symobjs: # delete pre existing objects
				vs.DelObject(obj)
	
		#outputs
		self.Params.symbol.value = symH