	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Extrude Path/Input/Path/EP Extrude Path Input UI.py;
	# Created by Bill Wood 29/7/18
	# Edited 25/4/2020 Converted to VW 2020
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if (type(inputObject) == vs.Handle): # single object
				objlistH = [inputObject]
			else:
				objlistH = inputObject # list of objects??
			
			newobjlistH = []
			if isinstance(objlistH,list):
				for objH in objlistH:
					objtype = vs.GetTypeN(objH)			
					if objtype in [2,3,4,5,6,21,25,111]: # 2D/3D path objects
						newobjlistH.append(objH)
				if len(newobjlistH) == 0:
					return 0 # invalid objects in objectlist
				else:
					return newobjlistH
			else:
				return 0 # not a list
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "EP Extrude Path Input UI" )
		this.SetListAbsorb()			  # mark this node as list absorbing
		this.SetDescription( 'Extrude path options definition.' )
	
		#Input Ports
		pathpioH = Marionette.PortIn(vs.Handle(0), 'pathH(s)')
		pathpioH.SetDescription( "Path object used for extrusion" )
	
		#OIP Input Controls
		startoffset = Marionette.OIPControl('Start Offset', Marionette.WidgetType.RealCoord, 0)
		startoffset.SetDescription( "Path offset at start" )
	
		endoffset = Marionette.OIPControl('End Offset', Marionette.WidgetType.RealCoord, 0)
		endoffset.SetDescription( "Path offset at end" )
		
		bulge = Marionette.OIPControl('Curve Bulge (EL)', Marionette.WidgetType.RealCoord, 0)
		bulge.SetDescription( "Split path at vertex ends (EP Line only)" )
	
		splitvertex = Marionette.OIPControl('Split at all end vertices (EPP)', Marionette.WidgetType.Bool, False)
		splitvertex.SetDescription( "Split path at vertex ends (EP Planar only)" )
	
		reversepath = Marionette.OIPControl('Reverse Path', Marionette.WidgetType.Bool, False)
		reversepath.SetDescription( "Reverse direction of path object" )
	
		showpath = Marionette.OIPControl('Show Path', Marionette.WidgetType.Bool, False)
		showpath.SetDescription( "Show path object route" )
	
		#Output Ports
		shape = Marionette.PortOut('Options')
		shape.SetDescription( "The resultant path options definition" )
	
	#BEHAVIOR
	
	def RunNode(self):
		#inputs
		pathpioH = self.Params.getinputObject(self.Params.pathpioH.value)
		if pathpioH:
			startoffset  = self.Params.startoffset.value
			endoffset	 = self.Params.endoffset.value
			bulge		 = self.Params.bulge.value
			splitvertex  = self.Params.splitvertex.value
			reversepath  = self.Params.reversepath.value
			showpath	 = self.Params.showpath.value
	
			options = {'PI':[pathpioH,{'startoffset':str(startoffset),'endoffset':str(endoffset),'reversepath':str(bool(reversepath)),'showloci':str(bool(showpath)),'splitatvertex':str(bool(splitvertex)),'bulge':str(bulge)}]}
		else:
			options = 0
	
		#outputs
		self.Params.shape.value = options
	