	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Extrude Path/Input/Path/EP Extrude Path Input.py;
	# Created by Bill Wood 29/7/18
	# Edited 25/4/2020 Converted to VW 2020
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if (type(inputObject) == vs.Handle): # single object
				objlistH = [inputObject]
			else:
				objlistH = inputObject # list of objects??
			
			newobjlistH = []
			if isinstance(objlistH,list):
				for objH in objlistH:
					objtype = vs.GetTypeN(objH)			
					if objtype in [2,3,4,5,6,21,25,111]: # 2D/3D path objects
						newobjlistH.append(objH)
				if len(newobjlistH) == 0:
					return 0 # invalid objects in objectlist
				else:
					return newobjlistH
			else:
				return 0 # not a list
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "EP Extrude Path Input" )
		this.SetListAbsorb()			  # mark this node as list absorbing
		this.SetDescription( 'The EP Extrude Path input definition.' )
	
		#Input Ports
		pathpioH = Marionette.PortIn(vs.Handle(0), 'hPath(s)')
		pathpioH.SetDescription( "Path object used for extrusion" )
		
		startoffset = Marionette.PortIn(0, 'nStartOffset')
		startoffset.SetDescription( "Path offset at start" )
	
		endoffset = Marionette.PortIn(0, 'nEndOffset')
		endoffset.SetDescription( "Path offset at end" )
	
		reversepath = Marionette.PortIn(False, 'bReversePath')
		reversepath.SetDescription( "Reverse direction of path object" )
	
		showpath = Marionette.PortIn(False, 'bShowPath')
		showpath.SetDescription( "Show path object route" )
	
		splitvertex = Marionette.PortIn(False, 'bSplitPath (EP)')
		splitvertex.SetDescription( "Split path at vertex ends (EP Path 3D and Planar only)" )
	
		bulge = Marionette.PortIn(0, 'nCurveBulge (EL)')
		bulge.SetDescription( "Arc bulge at midpoint (EP Linear only)" )
	
	
		#Output Ports
		shape = Marionette.PortOut('Options')
		shape.SetDescription( "The resultant path options definition" )
	
	#BEHAVIOR
	
	def RunNode(self):
		def getnumbval(value,defvalue):
			if isinstance(value,(list,tuple)):
				value = value[0]
			if isinstance(value,(int,float)):
				return value
			else:
				return defvalue
	
		def getboolval(bvalue):
			if isinstance(bvalue,list):
				return bvalue[0]
			else:
				return bvalue
	
		#inputs
		pathpioH = self.Params.getinputObject(self.Params.pathpioH.value)
		if pathpioH:
			startoffset  = getnumbval(self.Params.startoffset.value,0)
			endoffset	 = getnumbval(self.Params.endoffset.value,0)
			reversepath  = getboolval(self.Params.reversepath.value)
			showpath	 = getboolval(self.Params.showpath.value)
			splitvertex  = getboolval(self.Params.splitvertex.value)
			bulge		 = getnumbval(self.Params.bulge.value,0) # PIO will set minimum value
	
			options = {'PI':[pathpioH,{'startoffset':str(startoffset),'endoffset':str(endoffset),'reversepath':str(bool(reversepath)),'showloci':str(bool(showpath)),'splitatvertex':str(bool(splitvertex)),'bulge':str(bulge)}]}
		else:
			options = 0
	
		#outputs
		self.Params.shape.value = options