	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Extrude Path/Input/Spacing/EP Extrude Spacing Input.py;
	# Created by Bill Wood 29/7/18
	# Edited Converted to VW 2020 25/4/2020
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if (type(inputObject) == vs.Handle): # single object
				objlistH = [inputObject]
			else:
				objlistH = inputObject # list of objects??
			
			newobjlistH = []
			if isinstance(objlistH,list):
				for objH in objlistH:
					objtype = vs.GetTypeN(objH)			
					if objtype in [2,3,4,5,6,21,25,111]: # 2D/3D path objects
						newobjlistH.append(objH)
				if len(newobjlistH) == 0:
					return 0 # invalid objects in objectlist
				else:
					return newobjlistH
			else:
				return 0 # not a list
	
		def getsetout(value):
			return value
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "EP Extrude Spacing Input" )
		this.SetListAbsorb()			  # mark this node as list absorbing
		this.SetDescription( 'The EP Extrude Spacing input definition.' )
	
		#Input Ports
		pathpioH = Marionette.PortIn(vs.Handle(0), 'hPath(s)')
		pathpioH.SetDescription( "Path object used for extrusion" )
		
		spacing = Marionette.PortIn(0, 'nSpacing')
		spacing.SetDescription( "Spacing distance along path" )
	
		setout = Marionette.PortIn(0, 'Setout')
		setout.SetDescription( "Setout method used for plaving extrudes" )
	
		shapeinplan = Marionette.PortIn(False, 'bShapeinPlan')
		shapeinplan.SetDescription( "Extrude shape in plan" )
	
		endsfollowpath = Marionette.PortIn(False, 'bFollowPath')
		endsfollowpath.SetDescription( "Extrude shapes to follow path" )
	
		extrudeby = Marionette.PortIn(0, 'nExtrudehgt')
		extrudeby.SetDescription( "Extrude height" )
	
		extrudeoffset = Marionette.PortIn(0, 'nExtrudeOffset')
		extrudeoffset.SetDescription( "Extrude offset" )
	
		startoffset = Marionette.PortIn(0, 'nStartOffset')
		startoffset.SetDescription( "Path offset at start" )
	
		endoffset = Marionette.PortIn(0, 'nEndOffset')
		endoffset.SetDescription( "Path offset at end" )
	
		toprotation = Marionette.PortIn(0, 'nTopRotation')
		toprotation.SetDescription( "Top rotation" )
	
		siderotation = Marionette.PortIn(0, 'nSideRotation')
		siderotation.SetDescription( "Side rotation" )
	
		toprotfollowspath = Marionette.PortIn(False, 'bRotationfollowsPath')
		toprotfollowspath.SetDescription( "Top rotation follows path" )
	
		splitvertex = Marionette.PortIn(False, 'bSplitPath')
		splitvertex.SetDescription( "Split path at vertex ends" )
	
		reversepath = Marionette.PortIn(False, 'bReversePath')
		reversepath.SetDescription( "Reverse direction of path object" )
	
		showpath = Marionette.PortIn(False, 'bShowPath')
		showpath.SetDescription( "Show path object route" )
	
	
		#Output Ports
		shape = Marionette.PortOut('Options')
		shape.SetDescription( "The resultant path options definition" )
	
	#BEHAVIOR
	
	def RunNode(self):
		def getnumbval(value,defvalue):
			if isinstance(value,(list,tuple)):
				value = value[0]
			if isinstance(value,(int,float)):
				return value
			else:
				return defvalue
	
		def getboolval(bvalue):
			if isinstance(bvalue,list):
				return bvalue[0]
			else:
				return bvalue
	
		#inputs
		pathpioH = self.Params.getinputObject(self.Params.pathpioH.value)
		if pathpioH:	 
			setout   = self.Params.getsetout(self.Params.setout.value)
			if isinstance(setout,dict) and ('SS' in setout):
				setout = setout['SS']
			else:
				setout = 'Equal' # default
	
			spacing			= getnumbval(self.Params.spacing.value,0)
			extrudeby		= getnumbval(self.Params.extrudeby.value,0)
			shapeinplan		= getboolval(self.Params.shapeinplan.value)
			endsfollowpath	= getboolval(self.Params.endsfollowpath.value)
			extrudeoffset	= getnumbval(self.Params.extrudeoffset.value,0)
			startoffset		= getnumbval(self.Params.startoffset.value,0)
			endoffset		= getnumbval(self.Params.endoffset.value,0)
			toprotation		= getnumbval(self.Params.toprotation.value,0)
			siderotation	= getnumbval(self.Params.siderotation.value,0)
			toprotfollowspath = getboolval(self.Params.toprotfollowspath.value)
			reversepath		= getboolval(self.Params.reversepath.value)
			showpath		= getboolval(self.Params.showpath.value)
			splitvertex		= getboolval(self.Params.splitvertex.value)
	
			options = {'SI':[pathpioH,{'spacing':str(spacing),
			'setouttype': setout,
			'extrudeby':str(extrudeby),
			'extrudeoffset':str(extrudeoffset),
			'startoffset':str(startoffset),
			'endoffset':str(endoffset),
			'toprotation':str(toprotation),
			'siderotation':str(siderotation),
			'siderotation':str(siderotation),
			'useplanshape':str(bool(shapeinplan)),
			'extrudefollowpath':str(bool(endsfollowpath)),
			'rotationfollowspath':str(bool(toprotfollowspath)),
			'reversepath':str(bool(reversepath)),
			'showloci':str(bool(showpath)),
			'splitatvertex':str(bool(splitvertex))}]}
		else:
			options = 0
	
		#outputs
		self.Params.shape.value = options