	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Extrude Path/Shapes/Bulb Flat/EP Bulb Flat Custom.py;
	# Created by Bill Wood 29/7/18
	
	class Params(metaclass = Marionette.OrderedClass):
	#APPEARANCE
		nodename = "EP Bulb Flat Custom"
		nodedesc = 'Set a custom Bulb Flat section size'
	
		def getsizedeffirst(filename):
			import os
			fref = os.path.join(vs.GetFolderPath(20) , filename )
			content = []
			contentvals = []
			if os.path.exists(fref):
	
				f = open(fref)
				upi  = f.readline()
				line = f.readline()
				line  = line.splitlines() # omit line break
				sline = line[0].split('\t')
				sizeref = sline.pop(0)
	
				f.close()
			else:
				sizeref = 0
			return sizeref
	
		def getsectionstring(self,b,t,c,d,rb,rf,theta,sharpCorners):
			nodetype   = 'Bulb Flat'
			shapetype  = 'Shape Plugin'
			sep = chr(92) # backslash
	
			impunits = ['Inches','Feet','Feet & Inches','Yards','Miles']
			if vs.GetPrefString(153) in impunits:
				nodeseries = 'BSI'
				filename = 'StructShape_BulbFlat-BSI.txt'
			else:
				nodeseries = 'DIN'
				filename = 'StructShape_BulbFlat-DIN.txt'
	
			params = '\\b\\t\\c\\d\\rb\\rf\\theta\\'
			paramvals = sep+str(b)+sep+str(t)+sep+str(c)+sep+str(d)+sep+str(rb)+sep+str(rf)+sep+str(theta)+sep
			upi = vs.GetPrefReal(152)
	
			# EP Popup references
			series = ['Metric','Inch','BSI','DIN','JIS','ANZ']
			funcs  = ['WFlange','IBeam','Channel','Angle','Tee','SqTubing','RectTubing','RoundTubing','BulbFlat','ZSection']
			types  = ['Wide Flange','I-Beam','Channel','Angle','Tee','Square Tubing','Rectangular Tubing','Round Tubing','Bulb Flat','Z-Section']
	
			# <pioclass = 1>_<pioname>_<piotype>_<pioname filref>_<seriesno>_<seriesname>_<sizeno>_<sizename>_<customset 0 = FALSE 1 = TRUE>_
			typeref = types.index(nodetype)
			funcref = funcs[typeref]
	
			seriesref = series.index(nodeseries)
			seriesname = series[seriesref]
			
			sizename = self.Params.getsizedeffirst(filename)
	
			if sizename:
				# Example  _1_Wide Flange_1_WFlange_1_Metric_268_W100 x 19.3_0_
				shapedef = nodetype + ' ' + seriesname + ' Custom'
				piodef = sep +'1'+ sep + types[typeref] + sep + str(typeref+1) + sep + funcs[typeref] + sep + str(seriesref+1) + sep + seriesname + sep + '1' + sep + sizename + sep + '1' + sep # last 1 is custom size
				nodedefinition = {'IS':{'shapetype':shapetype,'shapedef':shapedef,'__pioname':nodetype,'__piodef':piodef,'__sharpcorners':str(bool(sharpCorners)),'__noholes':'True','__upi':upi,'__params':params,'__paramvals':paramvals}}
			else:
				nodedefinition = 0
	
			return nodedefinition
		
		#Name
		this = Marionette.Node( nodename )
		this.SetDescription( nodedesc )
		
		#OIP Controls
		
		b = Marionette.OIPControl( 'Width', Marionette.WidgetType.RealCoord, 120.0)
		b.SetDescription('The overall vertical height of blub flat')
	
		t = Marionette.OIPControl( 'Thickness', Marionette.WidgetType.RealCoord, 6.0)
		t.SetDescription('The thickness of vertical leg of blub flat')
	
		c = Marionette.OIPControl( 'Bulb Height', Marionette.WidgetType.RealCoord, 17.0)
		c.SetDescription('The overall vertical height of bulb')
	
		d = Marionette.OIPControl( 'Bulb Width', Marionette.WidgetType.RealCoord, 17.7)
		d.SetDescription('The horizontal width of bulb')
	
		rb = Marionette.OIPControl( 'Bulb Radius', Marionette.WidgetType.RealCoord, 5.0)
		rb.SetDescription('The bulb radius between top and splay edges')
	
		rf = Marionette.OIPControl( 'Fillet Radius', Marionette.WidgetType.RealCoord, 1.5)
		rf.SetDescription('The fillet radius at all vertices other than the bulb')
		
		theta = Marionette.OIPControl( 'Bulb Angle (Deg)', Marionette.WidgetType.RealCoord, 30.0)
		theta.SetDescription('The angle splay of bulb')
	
		sharpCorners = Marionette.OIPControl( 'Sharp corners', Marionette.WidgetType.Bool, False)
		sharpCorners.SetDescription( "Section shape to replace fillets with sharp corners" )
	
		#Output Ports
		sectdef = Marionette.PortOut('Shape')
		sectdef.SetDescription( "The resultant shape definition" )
	
	#BEHAVIOR
	
	def RunNode(self):
		#inputs
		b	= self.Params.b.value
		t	= self.Params.t.value
		c	= self.Params.c.value
		d  = self.Params.d.value
		rb  = self.Params.rb.value
		rf   = self.Params.rf.value
		theta   = self.Params.theta.value
	
		valuetest = ((b > 0) and (t > 0) and (c > 0) and (d > 0) and (rb >= 0) and (rf >= 0) and (theta >= 0))
		if valuetest:
			sharpCorners = self.Params.sharpCorners.value
	
			#outputs
			sectionstring = self.Params.getsectionstring(self,b,t,c,d,rb,rf,theta,sharpCorners)
		else:
			sectionstring = 0
	
		self.Params.sectdef.value = sectionstring