	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Extrude Path/Shapes/Tee/EP Tee Custom.py;
	# Created by Bill Wood 29/7/18
	
	class Params(metaclass = Marionette.OrderedClass):
	#APPEARANCE
		nodename = "EP Tee Custom"
		nodedesc = 'Set a custom Tee section size'
	
		def getsizedeffirst(filename):
			import os
			fref = os.path.join(vs.GetFolderPath(20) , filename )
			content = []
			contentvals = []
			if os.path.exists(fref):
	
				f = open(fref)
				upi  = f.readline()
				line = f.readline()
				line  = line.splitlines() # omit line break
				sline = line[0].split('\t')
				sizeref = sline.pop(0)
	
				f.close()
			else:
				sizeref = 0
			return sizeref
	
		def getsectionstring(self,d,w,tw,tf,wslope,fslope,rtoe,rflange,rfillet,sharpCorners):
			nodetype   = 'Tee'
			shapetype  = 'Shape Plugin'
			sep = chr(92) # backslash
	
			impunits = ['Inches','Feet','Feet & Inches','Yards','Miles']
			if vs.GetPrefString(153) in impunits:
				nodeseries = 'Inch'
				filename = 'StructShape_Tee-Inch.txt'
			else:
				nodeseries = 'Metric'
				filename = 'StructShape_Tee-Metric.txt'
	
			params = '\\d\\w\\tw\\tf\\slope\\slope_flange\\r_Toe\\rt\\rf\\'
			paramvals = sep+str(d)+sep+str(w)+sep+str(tw)+sep+str(tf)+sep+str(wslope)+sep+str(fslope)+sep+str(rtoe)+sep+str(rflange)+sep+str(rfillet)+sep
			upi = vs.GetPrefReal(152)
	
			# EP Popup references
			series = ['Metric','Inch','BSI','DIN','JIS','ANZ']
			funcs  = ['WFlange','IBeam','Channel','Angle','Tee','SqTubing','RectTubing','RoundTubing','BulbFlat','ZSection']
			types  = ['Wide Flange','I-Beam','Channel','Angle','Tee','Square Tubing','Rectangular Tubing','Round Tubing','Bulb Flat','Z-Section']
	
			# <pioclass = 1>_<pioname>_<piotype>_<pioname filref>_<seriesno>_<seriesname>_<sizeno>_<sizename>_<customset 0 = FALSE 1 = TRUE>_
			typeref = types.index(nodetype)
			funcref = funcs[typeref]
	
			seriesref = series.index(nodeseries)
			seriesname = series[seriesref]
			
			sizename = self.Params.getsizedeffirst(filename)
	
			if sizename:
				# Example  _1_Wide Flange_1_WFlange_1_Metric_268_W100 x 19.3_0_
				shapedef = nodetype + ' ' + seriesname + ' Custom'
				piodef = sep +'1'+ sep + types[typeref] + sep + str(typeref+1) + sep + funcs[typeref] + sep + str(seriesref+1) + sep + seriesname + sep + '1' + sep + sizename + sep + '1' + sep # last 1 is custom size
				nodedefinition = {'IS':{'shapetype':shapetype,'shapedef':shapedef,'__pioname':nodetype,'__piodef':piodef,'__sharpcorners':str(bool(sharpCorners)),'__noholes':'True','__upi':upi,'__params':params,'__paramvals':paramvals}}
			else:
				nodedefinition = 0
	
			return nodedefinition
		
		#Name
		this = Marionette.Node( nodename )
		this.SetDescription( nodedesc )
		
		#OIP Controls
		d = Marionette.OIPControl( 'Depth', Marionette.WidgetType.RealCoord, 76.581)
		d.SetDescription('The depth of tee')
	
		w = Marionette.OIPControl( 'Flange Width', Marionette.WidgetType.RealCoord, 101.6)
		w.SetDescription('The width of tee flange')
	
		tw = Marionette.OIPControl( 'Web Thickness', Marionette.WidgetType.RealCoord, 5.842)
		tw.SetDescription('The thickness of tee web')
	
		tf = Marionette.OIPControl( 'Flange Thickness', Marionette.WidgetType.RealCoord, 7.112)
		tf.SetDescription('The thickness of tee flange')
		
		wslope = Marionette.OIPControl( 'Web Slope', Marionette.WidgetType.RealCoord, 0.0)
		wslope.SetDescription('The angle of web from vertical')
		
		fslope = Marionette.OIPControl( 'Flange Slope', Marionette.WidgetType.RealCoord, 0.0)
		fslope.SetDescription('The angle of flange from horizontal')
	
		rtoe = Marionette.OIPControl( 'Toe Radius', Marionette.WidgetType.RealCoord, 0.0)
		rtoe.SetDescription('The toe radius of vertical web')
	
		rflange = Marionette.OIPControl( 'Flange Radius', Marionette.WidgetType.RealCoord, 0.0)
		rflange.SetDescription('The flange radius of the ends of flanges')
	
		rfillet = Marionette.OIPControl( 'Fillet Radius', Marionette.WidgetType.RealCoord, 6.35)
		rfillet.SetDescription('The fillet radius between flange and web')
	
		sharpCorners = Marionette.OIPControl( 'Sharp corners', Marionette.WidgetType.Bool, False)
		sharpCorners.SetDescription( "Section shape to replace fillets with sharp corners" )
	
		#Output Ports
		sectdef = Marionette.PortOut('Shape')
		sectdef.SetDescription( "The resultant shape definition" )
	
	#BEHAVIOR
	
	def RunNode(self):
		#inputs
		d	   = self.Params.d.value
		w	   = self.Params.w.value
		tw	   = self.Params.tw.value
		tf	   = self.Params.tf.value
		wslope = self.Params.wslope.value
		fslope  = self.Params.fslope.value
		rtoe	= self.Params.rtoe.value
		rflange	= self.Params.rflange.value
		rfillet	= self.Params.rfillet.value
	
		valuetest = ((d > 0) and (w > 0) and (tw > 0) and (tf > 0) and (wslope >= 0) and (fslope >= 0) and (rtoe >= 0) and (rflange >= 0) and (rfillet >= 0))
		if valuetest:
			sharpCorners = self.Params.sharpCorners.value
	
			#outputs
			sectionstring = self.Params.getsectionstring(self,d,w,tw,tf,wslope,fslope,rtoe,rflange,rfillet,sharpCorners)
		else:
			sectionstring = 0
	
		self.Params.sectdef.value = sectionstring