	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Extrude Path/Shapes/Z Section/EP Z Section Custom.py;
	# Created by Bill Wood 29/7/18
	
	class Params(metaclass = Marionette.OrderedClass):
	#APPEARANCE
		nodename = "EP Z Section Custom"
		nodedesc = 'Set a custom Z section size'
	
		def getsizedeffirst(filename):
			import os
			fref = os.path.join(vs.GetFolderPath(20) , filename )
			content = []
			contentvals = []
			if os.path.exists(fref):
	
				f = open(fref)
				upi  = f.readline()
				line = f.readline()
				line  = line.splitlines() # omit line break
				sline = line[0].split('\t')
				sizeref = sline.pop(0)
	
				f.close()
			else:
				sizeref = 0
			return sizeref
	
		def getsectionstring(self,d,a,b,tw,t,rtA,rtB,rf,sharpCorners,noholes):
			nodetype   = 'Z-Section'
			shapetype  = 'Shape Plugin'
			sep = chr(92) # backslash
	
			impunits = ['Inches','Feet','Feet & Inches','Yards','Miles']
			if vs.GetPrefString(153) in impunits:
				nodeseries = 'Inch'
				filename = 'StructShape_ZSection-DIN.txt'
			else:
				nodeseries = 'Metric'
				filename = 'StructShape_ZSection-Metric.txt'
	
			params = '\\d\\a\\b\\tw\\t\\rtA\\rtB\\rf\\'
			paramvals = sep+str(d)+sep+str(a)+sep+str(b)+sep+str(tw)+sep+str(t)+sep+str(rtA)+sep+str(rtB)+sep+str(rf)+sep
			upi = vs.GetPrefReal(152)
	
			# EP Popup references
			series = ['Metric','Inch','BSI','DIN','JIS','ANZ']
			funcs  = ['WFlange','IBeam','Channel','Angle','Tee','SqTubing','RectTubing','RoundTubing','BulbFlat','ZSection']
			types  = ['Wide Flange','I-Beam','Channel','Angle','Tee','Square Tubing','Rectangular Tubing','Round Tubing','Bulb Flat','Z-Section']
	
			# <pioclass = 1>_<pioname>_<piotype>_<pioname filref>_<seriesno>_<seriesname>_<sizeno>_<sizename>_<customset 0 = FALSE 1 = TRUE>_
			typeref = types.index(nodetype)
			funcref = funcs[typeref]
	
			seriesref = series.index(nodeseries)
			seriesname = series[seriesref]
			
			sizename = self.Params.getsizedeffirst(filename)
	
			if sizename:
				# Example  _1_Wide Flange_1_WFlange_1_Metric_268_W100 x 19.3_0_
				shapedef = nodetype + ' ' + seriesname + ' Custom'
				piodef = sep +'1'+ sep + types[typeref] + sep + str(typeref+1) + sep + funcs[typeref] + sep + str(seriesref+1) + sep + seriesname + sep + '1' + sep + sizename + sep + '1' + sep # last 1 is custom size
				nodedefinition = {'IS':{'shapetype':shapetype,'shapedef':shapedef,'__pioname':nodetype,'__piodef':piodef,'__sharpcorners':str(bool(sharpCorners)),'__noholes':str(bool(noholes)),'__upi':upi,'__params':params,'__paramvals':paramvals}}
			else:
				nodedefinition = 0
	
			return nodedefinition
		
		#Name
		this = Marionette.Node( nodename )
		this.SetDescription( nodedesc )
		
		#OIP Controls
		d = Marionette.OIPControl( 'Depth', Marionette.WidgetType.RealCoord, 100.0)
		d.SetDescription('The depth of Z section')
	
		a = Marionette.OIPControl( 'Flange Width (Top)', Marionette.WidgetType.RealCoord, 55.0)
		a.SetDescription('The top flange width of Z section')
	
		b = Marionette.OIPControl( 'Flange Width (Bottom)', Marionette.WidgetType.RealCoord, 55.0)
		b.SetDescription('The bottom flange width of Z section')
	
		tw = Marionette.OIPControl( 'Web Thickness', Marionette.WidgetType.RealCoord, 6.5)
		tw.SetDescription('The web thickness of Z section')
	
		t = Marionette.OIPControl( 'Flange Thickness', Marionette.WidgetType.RealCoord, 8.0)
		t.SetDescription('The flange thickness of Z section')
	
		rtA = Marionette.OIPControl( 'Toe Radius (Top)', Marionette.WidgetType.RealCoord, 4.0)
		rtA.SetDescription('The top toe radius of Z section')
	
		rtB = Marionette.OIPControl( 'Toe Radius (Bottom)', Marionette.WidgetType.RealCoord, 4.0)
		rtB.SetDescription('The bottom toe radius of Z section')
	
		rf = Marionette.OIPControl( 'Fillet Radius', Marionette.WidgetType.RealCoord, 8.0)
		rf.SetDescription('The fillet radius between flange and web')
	
		sharpCorners = Marionette.OIPControl( 'Sharp corners', Marionette.WidgetType.Bool, False)
		sharpCorners.SetDescription( "Section shape to replace fillets with sharp corners" )
	
		#Output Ports
		sectdef = Marionette.PortOut('Shape')
		sectdef.SetDescription( "The resultant shape definition" )
	
	#BEHAVIOR
	
	def RunNode(self):
		#inputs
		d  = self.Params.d.value
		a  = self.Params.a.value
		b  = self.Params.b.value
		tw = self.Params.tw.value
		t  = self.Params.t.value
		rtA = self.Params.rtA.value
		rtB = self.Params.rtB.value
		rf = self.Params.rf.value
	
		valuetest = ((d > 0) and (a > 0) and (b > 0) and (tw > 0) and (t > 0) and (rtA >= 0) and (rtB >= 0) and (rf >= 0))
		if valuetest:
			sharpCorners = self.Params.sharpCorners.value
	
			#outputs
			sectionstring = self.Params.getsectionstring(self,d,a,b,tw,t,rtA,rtB,rf,sharpCorners,False)
		else:
			sectionstring = 0
	
		self.Params.sectdef.value = sectionstring