	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Extrude Path/Tool/EP Extrude.py;
	# Created by Bill Wood 29/7/18
	# Edited
	# Converted to VW 2020 25/4/20
	# Rountine createfrom3Dpoly changed 24/2/21
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if isinstance(inputObject,list):
				inputObject = inputObject[0]
	
			if isinstance(inputObject,dict):
				if 'EI' in inputObject: # Extrude Shape
				# {'EI':[location,{'topoffset':,'extrudehgt':,'baseZ':,'skewX':,'skewY':,'verticaloffsets':,'maintainshape':}]}
					content = inputObject['EI']
					location = content[0]
					params   = content[1]
	
					if ('EO' in location): # {'EO':inputObject}
						objlistH   = location['EO']
						result = [0,objlistH,params]
					elif ('EP' in location): #{'EP':[inputPosition,WProtations]}
						temp = location['EP']
						positions  = temp[0]
						rotation = temp[1]
						result = [1,positions,rotation,params] # pos is a list defining origin, WP set to world plane
					elif ('EL' in location): #{'EL':[(inputFrom,inputTo),(rotX, rotY, rotZ)]}
						temp = location['EL']
						linefromtos  = temp[0]
						rotation = temp[1]
						result = [2,linefromtos,rotation,params] # pos is a list defining origin, WP set to world plane
					else:
						result = 0 # default fail
	
				elif 'PI' in inputObject: # Extrude Path
				# {'PI':[pathpioH,{'startoffset':,'endoffset','reversepath','showloci':,'splitatvertex':,'bulge':}]}
					content = inputObject['PI']
					objlistH = content[0]
					params = content[1]
	
					result = [0,objlistH,params]
	
				elif 'SI' in inputObject: # Extrude Spacing
				# {'SI':[pathpioH,{'spacing':,'setouttype':,'extrudeby':,'extrudeoffset':,'startoffset':,'endoffset':,'toprotation':,'siderotation':,'siderotation':,'useplanshape':,'extrudefollowpath':,'toprotfollowspath':,'reversepath':,'showloci':,'splitatvertex':}]}
					content = inputObject['SI']
					objlistH = content[0]
					params = content[1]
	
					result = [3,objlistH,params]
	
				else:
					result = 0 # default fail
			else: # default values
				result = 0 # default fail
		
			return result
	
		def getinputShape(inputShape):
			result = {'shapedef':'Standard','shapetype':'Rectangle','__standardshape':'1', 'val1':'10mm','val2':'10mm','thickness':'0'} # Default
			if isinstance(inputShape,list):
				inputShape = inputShape[0]
			if isinstance(inputShape,dict):
				if 'IS' in inputShape:
					result = inputShape['IS']
					
			return result
		
		def getclass(classname):
			if isinstance(classname,list):
				classname = classname[0] # choose first entry
	
			if isinstance(classname,str):
				return classname
			else:
				return 'None'
	
		def getorientation(orientation):
			result = {'xoffset':'0','yoffset':'0','shapeangle':'0','fliphori':'False','flipvert':'False','shapepivot':'As Shape'} # default
			if isinstance(orientation,list):
			 	orientation = orientation[0]
			if isinstance(orientation,dict):
				if 'SO' in orientation:
					result = orientation['SO']
	
			return result
	
		def getmitres(mitres):
			result = {'MI':{'startmitre':'None','endmitre':'None'}} # Default
			if isinstance(mitres,list):
				mitres = mitres[0]
			if isinstance(mitres,dict):
				if 'MI' in mitres:
					result = mitres['MI']
	
			return result
		
		def drawextrudebyposition(inputObject,inputShape,orientation,mitres,classname):
			# Single position definition
			# [1,pos,wprotation,params]
			del(inputObject[0]) # Remove first entry
	
			positions  = inputObject[0]
			wprotation = inputObject[1]
			
			params = inputObject[2]
			params.update(inputShape)
			params.update(orientation)
	
			extrudetype = 'Extrude Shape'
			for pos in positions:
				extrH = vs.CreateCustomObjectN(extrudetype,0,0,0,False)
				btest = vs.SetEntityMatrix(extrH ,pos,wprotation[0],wprotation[1],wprotation[2])
				for k in params:
					vs.SetRField(extrH,extrudetype,k,params[k])
	
				vs.SetClass(extrH, classname)
				status = vs.GetRField(extrH,extrudetype,'status')
			
			return (status,extrH)
			
		def drawextrudebyline(inputObject,inputShape,orientation,mitres,classname):
			# End position definition
			# [2,posfrom,posto,wprotation,params]
			del(inputObject[0]) # Remove first entry
			
			pos1 = inputObject[0]
			pos2 = inputObject[1]
			linelen = vs.Distance3D(pos1[0],pos1[1],pos1[2],pos2[0],pos2[1],pos2[2])
			# next line omited
			# linelen = 0
	
			wprotation = inputObject[2]
			xdir = wprotation[0]
			ydir = wprotation[1]
			zdir = wprotation[2]
	
			params = inputObject[3]
			params.update(inputShape)
			params.update(orientation)
			params.update(mitres)
			
			extrudetype = 'Extrude Linear'		
			extrH = vs.CreateCustomObjectN(extrudetype,0,0,0,False)
			if vs.SetEntityMatrixN(extrH,xdir,ydir,zdir,pos1):
				vs.SetRField(extrH,extrudetype,'LineLength',str(linelen))
				for k in params:
					vs.SetRField(extrH,extrudetype,k,params[k])
	
				vs.SetClass(extrH, classname)
				status = vs.GetRField(extrH,extrudetype,'status')
			else:
				status = 'Error in setting EP Linear orientation'
				vs.DelObject(extrH)
				extrH = 0;
	
			return (status,extrH)
	
		def drawextrudebyobject(inputObject,inputShape,orientation,mitres,classname):
			# Constants
			kSHAPEPIO   = 'Extrude Shape'
			kEXTRLINEAR = 'Extrude Linear'
			kEXTRPLANAR = 'Extrude Path Planar'
			kPATH3DPIO  = 'Extrude Path3D'
			kEXTRSPACEPLANAR = 'Extrude Spacing Planar'
			X = 0
			Y = 1
			Z = 2
			
			def getlinexaxis(lineH,rID):
				pt1 = vs.GetSegPt1(lineH)
				pt2 = vs.GetSegPt2(lineH)
				
				btest,loc = vs.PlanarPtTo3DModelPt(rID,pt1)
				btest,pt3D = vs.PlanarPtTo3DModelPt(rID,pt2)
				vX = (pt3D[X]-loc[X],pt3D[Y]-loc[Y],pt3D[Z]-loc[Z])
				return  (vX,loc)
				
			def getpolyvecs(pH,rID):
				pt2D = vs.GetPolyPt(pH,1)
				btest,p1 = vs.PlanarPtTo3DModelPt(rID,pt2D)
				
				pt2D = vs.GetPolyPt(pH,2)
				btest,p2 = vs.PlanarPtTo3DModelPt(rID,pt2D)
				
				pt2D = vs.GetPolyPt(pH,3)
				btest,p3 = vs.PlanarPtTo3DModelPt(rID,pt2D)
				
				vx = (p2[X]-p1[X],p2[Y]-p1[Y],p2[Z]-p1[Z])
				vy = (p3[X]-p1[X],p3[Y]-p1[Y],p3[Z]-p1[Z])
				return (vx,vy)
	
			# ================================== Creation routines
			def createfrom3Dpoly(srcpolyH,spacingset):
				X = 0
				Y = 1
				Z = 2
	
				def getmatrix(objH,pt1,pt2,cnt):
					v1 = (pt2[X]-pt1[X],pt2[Y]-pt1[Y],pt2[Z]-pt1[Z])
					if cnt >= 3:
						pt3 = vs.GetPolyPt3D(objH,2)
						v2 = (pt3[X]-pt1[X],pt3[Y]-pt1[Y],pt3[Z]-pt1[Z])
					else: #single line polygon
						v2 = vs.Perp(v1)
						if v2 == (0,0,0): # vertical
							v2 = (0,1,0)
	
					v3 = vs.CrossProduct(v1, v2)
					v2 = vs.CrossProduct(v1, v3)
					v2 = (-v2[X],-v2[Y],-v2[Z])
				
					mati = vs.UnitVec(v1)
					matj = vs.UnitVec(v2)
					matk = vs.UnitVec(v3)
					matk = (-matk[X],-matk[Y],-matk[Z]) # Reverse from solid conversion
				
					return (mati,matj,matk)
				
				def getpolypath(objH,pcnt,v1,v2):
					base = (v2[X]-v1[X],v2[Y]-v1[Y],v2[Z]-v1[Z])
					vs.ClosePoly()
					vs.BeginPoly()
					
					vs.AddPoint(0,0)
					vlen = vs.Norm(base)
					vs.AddPoint(vlen,0)
					
					for i in range(pcnt):
						v3 = vs.GetPolyPt3D(objH,i)
						pt  = (v3[X]-v1[X],v3[Y]-v1[Y],v3[Z]-v1[Z])
						ang = vs.AngBVec(base,pt)
						vlen = vs.Norm(pt)
						v3  = vs.Ang2Vec(ang,vlen)
						vs.AddPoint(v3[X],v3[Y])
						
					vs.EndPoly()
					
					return vs.LNewObj() #getpolypath
	
				# =========== createfrom3Dpoly
				pt1 = vs.GetPolyPt3D(srcpolyH,0)
				pt2 = vs.GetPolyPt3D(srcpolyH,1)
				cnt = vs.GetVertNum(srcpolyH)
	
				mat = getmatrix(srcpolyH,pt1,pt2,cnt)
				u = mat[0]
				v = mat[1]
				w = mat[2]
				
				pathH = getpolypath(srcpolyH,cnt,pt1,pt2)
				btest = vs.SetEntityMatrixN(pathH,u,v,w,pt1)
				
				if spacingset:
					extrudetype = kEXTRSPACEPLANAR
				else:
					extrudetype = kEXTRPLANAR
	
				extH = vs.CreateCustomObjectPath(extrudetype,pathH,None)
				
				return (extH,extrudetype)
				# createfrom3Dpoly}
	
			def createfromnurbspath(objH):
				pathH = vs.CreateDuplicateObject(objH,None)
				extH  = vs.CreateCustomObjectPath(kPATH3DPIO,pathH,None)
			
				return (extH,kPATH3DPIO)
			
				# createfromnurbspath
	
			def createfromloci(objH,loci3Dset):
				extH = vs.CreateCustomObject(kSHAPEPIO,0,0,0)
			
				if loci3Dset:
					pt3D = vs.GetLocus3D(objH) # always aligned with world axes
					vs.Move3DObj(extH,pt3D[X],pt3D[Y],pt3D[Z])
			
				else: # 2D loci
					if (vs.GetTypeN(objH) == 15):
						pt3D = vs.GetSymLoc3D(objH)
					else:
						pt2D = vs.GetLocPt(objH)
						refID = vs.GetPlanarRef(objH)
						btest,pt3D = vs.PlanarPtTo3DModelPt(refID,pt2D)
			
					btest,wpt3D,angx,angy,angz = vs.GetEntityMatrix(objH)
					btest = vs.SetEntityMatrix(extH,pt3D,angx,angy,angz)
	
				return (extH,kSHAPEPIO)
	
				# createfromloci
	
			def createfromshape(objH,objty,spacingset):
				refID = vs.GetPlanarRef(objH)
				if (objty == 2): #line
					xaxisV,locV = getlinexaxis(objH,refID)
				
					vs.Poly(0,0,1,0,0,1)
					polyH = vs.LNewObj()
					btest,pt3D,angx,angy,angz = vs.GetEntityMatrix(objH)
					btest = vs.SetEntityMatrix(polyH,pt3D,angx,angy,angz)
			
					v1,v2 = getpolyvecs(polyH,refID)
					vs.DelObject(polyH)
			
					v3 = vs.CrossProduct(v2, v1)	
					v2 = vs.CrossProduct(v3,xaxisV)
					v3 = (-v3[X],-v3[Y],-v3[Z])
	
					extrudetype = kEXTRLINEAR
					extH = vs.CreateCustomObject(kEXTRLINEAR,0,0,0)
					vs.SetRField(extH,kEXTRLINEAR,'Linelength',vs.HLength(objH))
					btest = vs.SetEntityMatrixN(extH,xaxisV,v2,v3,locV)
				else:
					pathH = vs.MakePolyline(objH)
					# Note Holes handled from input to Path(s) via external node
					if spacingset:
						extrudetype = kEXTRSPACEPLANAR
					else:
						extrudetype = kEXTRPLANAR
	
					extH = vs.CreateCustomObjectPath(extrudetype,pathH,None)
					vs.ResetObject(extH)
			
				return (extH,extrudetype)
			
				# createfromshape
			
			# END ================================== Creation routines
	
	
			def drawextrude(objH,spacingset):
				objtype = vs.GetTypeN(objH)
				if (objtype == 25):
					(extH,extrudetype) = createfrom3Dpoly(objH,spacingset)
				elif (objtype == 111):
					(extH,extrudetype) = createfromnurbspath(objH)
				elif (objtype == 17):  #2D locus
					(extH,extrudetype) = createfromloci(objH,False)
				elif (objtype == 9):   #3D locus
					(extH,extrudetype) = createfromloci(objH,True)
				elif objtype in	[2,3,4,5,6,13,21]: #2D objects
					(extH,extrudetype) = createfromshape(objH,objtype,spacingset)
	
				return (extH,extrudetype)
	
			# =====================drawextrudebyobject(inputObject,inputShape,orientation,mitres,classname)
			# Single reference object or list of objects
			# [objlistH,params]
			spacingset = bool(inputObject[0]) # Non Spacing = 0 Spacing = 3
			objlistH = inputObject[1]
			params = inputObject[2]
			params.update(inputShape)
			params.update(orientation)
	
			extrlistH = []
			status = []
	
	#		if len(objlistH) > 1:
	#			vs.BeginGroup()
	
			for objH in objlistH:
				(extrH,extrudetype) = drawextrude(objH,spacingset)
				extrlistH.append(extrH)
				for k in params:
					vs.SetRField(extrH,extrudetype,k,params[k])
	
				if mitres and (extrudetype == 'Extrude Linear'):
					for k in mitres:
						vs.SetRField(extrH,extrudetype,k,mitres[k])
				
				vs.SetClass(extrH, classname)
				objstatus = vs.GetRField(extrH,extrudetype,'status')
				if objstatus != '':
					status.append(objstatus)
	
			if len(extrlistH) == 1:
				grpH = vs.GetParent(extrlistH[0])
				if vs.GetTypeN(grpH) == 11:
					vs.HUngroup(grpH)
				return (status,extrlistH[0]) # single extrude
			else:
				# if len(objlistH) > 1:
				#	vs.EndGroup()
				return (status,extrlistH)
				# return (status,(vs.LNewObj(),extrlistH)) # group plus list of extrudes
				
			# =====================drawextrudebyobject(inputObject,inputShape,orientation,classname)
		
	#APPEARANCE
		#Name
		this = Marionette.Node( "EP Extrude" )
		this.SetListAbsorb()			 #  mark this node as list absorbing
		this.SetDescription( 'Create an EP Extrude plugin object based on supplied definition' )
	
		#Input Ports
		inputObject = Marionette.PortIn(0, 'Options' )
		inputObject.SetDescription( "The input path,locus or single position definition(s) and options for the EP Extrude object(s)" )
	
		inputShape = Marionette.PortIn(0, 'Shape' )
		inputShape.SetDescription( "The input shape definition for the EP Extrude object" )
	
		orientation = Marionette.PortIn(0, 'Orientation' )
		orientation.SetDescription( "The shape orientation for the EP Extrude object" )
	
		mitredef = Marionette.PortIn(0, 'Mitres (EL)' )
		mitredef.SetDescription( "The end mitre definition for the EP Linear object (only)" )
		
		classname = Marionette.PortIn('None', 'sClass' )
		classname.SetDescription( "The shape classname used for the EP Extrude object" )
	
		#Output Ports
		extrH = Marionette.PortOut('hExtrude(s)')
		extrH.SetDescription( "The result EP Extrude object or grouped Extrude objects" )
	
		status = Marionette.PortOut('sStatus')
		status.SetDescription( "The status of the created EP Extrude object(s)" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
		#inputs
		data = self.Params.inputObject.value
		objectData = self.Params.getinputObject(data)
			
		shapeData   = self.Params.getinputShape(self.Params.inputShape.value)
		orientation = self.Params.getorientation(self.Params.orientation.value)
		mitres		= self.Params.getmitres(self.Params.mitredef.value)
	
		# get mitredef
	
		classname   = self.Params.getclass(self.Params.classname.value)
		status = 0
		
		if objectData:
			shapebymethod = objectData[0]
	
			if shapebymethod == 1: # Extrude Shape from position
				status,result = self.Params.drawextrudebyposition(objectData,shapeData,orientation,mitres,classname)		
			elif shapebymethod == 2: # Extrude Line from end positions
				status,result = self.Params.drawextrudebyline(objectData,shapeData,orientation,mitres,classname)
			else: # 0 = Extrude Path/Shape/3D 3 = Spacing from objects
				status,result = self.Params.drawextrudebyobject(objectData,shapeData,orientation,mitres,classname) # values must be error checked on input
	
		else:
			result = 0
			status = 'No valid path shapes/positions'
				
		#outputs
		self.Params.extrH.value = result
		self.Params.status.value = status