	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Panels/Panel Other/Panel UI.py;
	# Created by Bill Wood 25/4/20
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if (type(inputObject) == vs.Handle): # single object
				objlistH = [inputObject]
			else:
				objlistH = inputObject # list of objects??
			
			newobjlistH = []
			if isinstance(objlistH,list):
				for objH in objlistH:
					objtype = vs.GetTypeN(objH)
					if objtype in [3,4,5,6,13,21,25]: # 2D objects plus 3D poly
						# check for 2 vertex 2D polygons and omit
						if not (objtype in [21,25] and vs.GetVertNum(objH) == 2):
							newobjlistH.append(objH)
				if len(newobjlistH) == 0:
					return 0 # invalid objects in objectlist
				else:
					return newobjlistH
			else:
				return 0 # not a list
	
		def drawpanel(shapeH,classname,params):
	
			kEXTRPANEL = 'Extrude Panel Planar'
			X = 0
			Y = 1
			Z = 2
	
			# ================================== Creation routines
			def createfrom3Dpoly(srcpolyH,extrudetype):
	
				def getmatrix(objH):
					pt1 = vs.GetPolyPt3D(objH,0)
					pt2 = vs.GetPolyPt3D(objH,1)
					pt3,zval = vs.Get3DCntr(objH)
				
					v1 = (pt2[X]-pt1[X],pt2[Y]-pt1[Y],pt2[Z]-pt1[Z])
					v2 = (pt3[X]-pt1[X],pt3[Y]-pt1[Y],zval-pt1[Z])
					v3 = vs.CrossProduct(v1, v2)
					v2 = vs.CrossProduct(v1, v3)
					v2 = (-v2[X],-v2[Y],-v2[Z])
				
					mati = vs.UnitVec(v1)
					matj = vs.UnitVec(v2)
					matk = vs.UnitVec(v3)
				
					return(mati,matj,matk,pt1)
	
				def InversePointsTransform(objH,mat):
	
					def InversePointTransform(pt,mat):
						mati = mat[0]
						matj = mat[1]
						matk = mat[2]
						matoffset = mat[3]
					
						dOffsetx = pt[0] - matoffset[0]
						dOffsety = pt[1] - matoffset[1]
						dOffsetz = pt[2] - matoffset[2]
					
						resx =  dOffsetx * mati[0] + dOffsety * mati[1] + dOffsetz * mati[2]
						resy =  dOffsetx * matj[0] + dOffsety * matj[1] + dOffsetz * matj[2]
						resz =  dOffsetx * matk[0] + dOffsety * matk[1] + dOffsetz * matk[2]
	
						return (resx,resy,resz)
	
						# InversePointTransform
	
					cnt = vs.GetVertNum(objH)
					# btest = True
					vs.ClosePoly()
					vs.BeginPoly()
					for i in range(cnt):
						(v1x,v1y,v1z) = vs.GetPolyPt3D(objH,i)
						v1 = (v1x,v1y,v1z)
						pt = InversePointTransform(v1,mat)
						vs.AddPoint(pt[0],pt[1])
					vs.EndPoly()
	
					return vs.LNewObj()
	
					# InversePointsTransform
	
				# =========== createfrom3Dpoly
				mat = getmatrix(srcpolyH)
				u = mat[0]
				v = mat[1]
				w = mat[2]
				offset = mat[3]
	
				pathH = InversePointsTransform(srcpolyH,mat)
				extH = vs.CreateCustomObjectPath(extrudetype,pathH,None)
				btest = vs.SetEntityMatrixN(extH,u,v,w,offset)
		
				return (extH)
	
				# createfrom3Dpoly}
	
			def createfromshape(objH,objty,extrudetype):
				refID = vs.GetPlanarRef(objH)
				pathH = vs.MakePolyline(objH)
				extH = vs.CreateCustomObjectPath(extrudetype,pathH,None)
				vs.ResetObject(extH)
	
				return (extH)
			
				# createfromshape
	
				# END ================================== Creation routines
	
	
			# ==========================================drawpanel(shapeH,classname,params)
			extrudetype = kEXTRPANEL
			objtype = vs.GetTypeN(shapeH)
			# Note objects filtered in getinputObject
			if (objtype == 25): # 3D polygon
				(panelH) = createfrom3Dpoly(shapeH,extrudetype)
			else:
				(panelH) = createfromshape(shapeH,objtype,extrudetype)
	
			for k in params:
				vs.SetRField(panelH,extrudetype,k,params[k])
			vs.SetClass(panelH, classname)
	
			return panelH
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "Panel UI" )
		this.SetListAbsorb()			 #  mark this node as list absorbing
		this.SetDescription( 'Create standard panel configuration' )
	
		#Input Ports
		shapes = Marionette.PortIn(vs.Handle(0), 'ShapeH(s)')
		shapes.SetDescription( "Shapes that define single panel size" )
	
		#OIP Input Controls
		classname = Marionette.OIPControl( 'Class', Marionette.WidgetType.PopupClasses, 'None')
		classname.SetDescription( "List of classes" )
	
		sectiontype = Marionette.OIPControl('Panel Shape', Marionette.WidgetType.Popup, 0, ['Flat','Pyramidal','Curved','Cushion'] )
		sectiontype.SetDescription( "List of valid panel types" )
	
		solid = Marionette.OIPControl( 'Solid format', Marionette.WidgetType.Bool, False)
		solid.SetDescription( "Set panel with solid body" )
	
		bulge = Marionette.OIPControl( 'Bulge', Marionette.WidgetType.RealCoord, 50.0)
		bulge.SetDescription('Panel bulge at peak (minimum 50mm) not applicable to Flat')
	
		edgeoffset = Marionette.OIPControl( 'Edge Offset', Marionette.WidgetType.RealCoord, 0.0)
		edgeoffset.SetDescription('Panel edge offset')
	
		hgtoffset = Marionette.OIPControl( 'Height Offset', Marionette.WidgetType.RealCoord, 0.0)
		hgtoffset.SetDescription('Panel height offset')
	
		bezelangle = Marionette.OIPControl( 'Bezel Angle', Marionette.WidgetType.RealAngle, 0.0)
		bezelangle.SetDescription('Panel bezel angle (maximum 85deg) not applicable to Cushion')
	
		thickness = Marionette.OIPControl( 'Thickness', Marionette.WidgetType.RealCoord, 2.0)
		thickness.SetDescription('Panel thickness (minimum 2mm)')
	
		gasket = Marionette.OIPControl( 'With gaskets', Marionette.WidgetType.Bool, False)
		gasket.SetDescription( "Set panel with gaskets" )
	
		gasketthick = Marionette.OIPControl( 'Gasket Thickness', Marionette.WidgetType.RealCoord, 2.0)
		gasketthick.SetDescription('Panel gasket thickness (minimum 2mm)')
	
		gasketextend = Marionette.OIPControl( 'Gasket Extension', Marionette.WidgetType.RealCoord, 0.0)
		gasketextend.SetDescription('Panel gasket extension (greater than zero)')
	
		gasketclass = Marionette.OIPControl('Gasket Class', Marionette.WidgetType.Popup, 0, ['Panel-Gasket-Glass','Panel-Gasket-Solid','Panel-Gasket-1','Panel-Gasket-2','Panel-Gasket-3','Panel-Gasket-4','Panel-Gasket-5'] )
		gasketclass.SetDescription( "List of valid gasket classes" )
	
		flipsides = Marionette.OIPControl( 'Flip sides', Marionette.WidgetType.Bool, False)
		flipsides.SetDescription( "Draw panel on opposite side of shape not applicable to Cushion" )
	
		#Output Ports
		panels = Marionette.PortOut('PanelH(s)')
		panels.SetDescription( "The resultant panel(s)" )
	
		status = Marionette.PortOut('sStatus')
		status.SetDescription( "The status of the created Panel object(s)" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
		shapes = self.Params.getinputObject(self.Params.shapes.value)
		status = 0
		
		if shapes:
			classname = self.Params.classname.value
	
			paneldef = {'panelshape' : str(self.Params.sectiontype.value),
			'paneltype'  : 'Single Thickness', # for flat option
			'solid' : str(bool(self.Params.solid.value)),
			'bulge' : str(self.Params.bulge.value),
			'edgeoffset' : str(self.Params.edgeoffset.value),
			'hgtoffset'  : str(self.Params.hgtoffset.value),
			'bezelangle' : str(self.Params.bezelangle.value),
			'thickness'  : str(self.Params.thickness.value),
			'flipsides'  : str(bool(self.Params.flipsides.value)),
			'withgaskets' : str(bool(self.Params.gasket.value)),
			'gasketthick' :str(self.Params.gasketthick.value),
			'gasketextend':str(self.Params.gasketextend.value),
			'gasketclass' :str(self.Params.gasketclass.value)
			}
	
			proj = vs.GetProjection(vs.ActLayer())
			if proj: # Top plan set = 6 returns all 3D views as 0
				vs.DoMenuTextByName('Standard Views',2) # Top
	
			if len(shapes) == 1: # single shape
				panelpios = self.Params.drawpanel(shapes[0],classname,paneldef)
			else:
				panelpios = []
				status = []
				vs.BeginGroup()
				for shape in shapes: # list of shapes
					panelH = self.Params.drawpanel(shape,classname,paneldef)
					panelpios.append(panelH)
					objstatus = vs.GetRField(panelH,'Extrude Panel Planar','status')
					if objstatus != '':
						status.append(objstatus)
				vs.EndGroup()
				panelpios = (vs.LNewObj(),panelpios)
		else:
			panelpios = 0 # fail
			status = 'No valid path shapes'
	
		#outputs
		self.Params.panels.value = panelpios
		self.Params.status.value = status
	