	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Panels/Panel Other/Pyramid Panel.py;
	# Created by Bill Wood 25/4/20
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if (type(inputObject) == vs.Handle): # single object
				objlistH = [inputObject]
			else:
				objlistH = inputObject # list of objects??
			
			newobjlistH = []
			if isinstance(objlistH,list):
				for objH in objlistH:
					objtype = vs.GetTypeN(objH)
					if objtype in [3,4,5,6,13,21,25]: # 2D objects plus 3D poly
						# check for 2 vertex 2D polygons and omit
						if not (objtype in [21,25] and vs.GetVertNum(objH) == 2):
							newobjlistH.append(objH)
				if len(newobjlistH) == 0:
					return 0 # invalid objects in objectlist
				else:
					return newobjlistH
			else:
				return 0 # not a list
				
		def getclass(classname):
			if isinstance(classname,list): # select random class from list
				refno = random(0,len(list)-1)
				classname = list[refno]
	
			if isinstance(classname,str):
				return classname
			else:
				return 'None'
	
		def drawpanel(shapeH,classname,params):
	
			kEXTRPANEL = 'Extrude Panel Planar'
			X = 0
			Y = 1
			Z = 2
	
			# ================================== Creation routines
			def createfrom3Dpoly(srcpolyH,extrudetype):
	
				def getmatrix(objH):
					pt1 = vs.GetPolyPt3D(objH,0)
					pt2 = vs.GetPolyPt3D(objH,1)
					pt3,zval = vs.Get3DCntr(objH)
				
					v1 = (pt2[X]-pt1[X],pt2[Y]-pt1[Y],pt2[Z]-pt1[Z])
					v2 = (pt3[X]-pt1[X],pt3[Y]-pt1[Y],zval-pt1[Z])
					v3 = vs.CrossProduct(v1, v2)
					v2 = vs.CrossProduct(v1, v3)
					v2 = (-v2[X],-v2[Y],-v2[Z])
				
					mati = vs.UnitVec(v1)
					matj = vs.UnitVec(v2)
					matk = vs.UnitVec(v3)
				
					return(mati,matj,matk,pt1)
	
				def InversePointsTransform(objH,mat):
	
					def InversePointTransform(pt,mat):
						mati = mat[0]
						matj = mat[1]
						matk = mat[2]
						matoffset = mat[3]
					
						dOffsetx = pt[0] - matoffset[0]
						dOffsety = pt[1] - matoffset[1]
						dOffsetz = pt[2] - matoffset[2]
					
						resx =  dOffsetx * mati[0] + dOffsety * mati[1] + dOffsetz * mati[2]
						resy =  dOffsetx * matj[0] + dOffsety * matj[1] + dOffsetz * matj[2]
						resz =  dOffsetx * matk[0] + dOffsety * matk[1] + dOffsetz * matk[2]
	
						return (resx,resy,resz)
	
						# InversePointTransform
	
					cnt = vs.GetVertNum(objH)
					# btest = True
					vs.ClosePoly()
					vs.BeginPoly()
					for i in range(cnt):
						(v1x,v1y,v1z) = vs.GetPolyPt3D(objH,i)
						v1 = (v1x,v1y,v1z)
						pt = InversePointTransform(v1,mat)
						vs.AddPoint(pt[0],pt[1])
					vs.EndPoly()
	
					return vs.LNewObj()
	
					# InversePointsTransform
	
				# =========== createfrom3Dpoly
				mat = getmatrix(srcpolyH)
				u = mat[0]
				v = mat[1]
				w = mat[2]
				offset = mat[3]
	
				pathH = InversePointsTransform(srcpolyH,mat)
				extH = vs.CreateCustomObjectPath(extrudetype,pathH,None)
				btest = vs.SetEntityMatrixN(extH,u,v,w,offset)
		
				return (extH)
	
				# createfrom3Dpoly}
	
			def createfromshape(objH,objty,extrudetype):
				refID = vs.GetPlanarRef(objH)
				pathH = vs.MakePolyline(objH)
				extH = vs.CreateCustomObjectPath(extrudetype,pathH,None)
				vs.ResetObject(extH)
	
				return (extH)
			
				# createfromshape
	
				# END ================================== Creation routines
	
	
			# ==========================================drawpanel(shapeH,classname,params)
			extrudetype = kEXTRPANEL
			objtype = vs.GetTypeN(shapeH)
			# Note objects filtered in getinputObject
			if (objtype == 25): # 3D polygon
				(panelH) = createfrom3Dpoly(shapeH,extrudetype)
			else:
				(panelH) = createfromshape(shapeH,objtype,extrudetype)
	
			for k in params:
				vs.SetRField(panelH,extrudetype,k,params[k])
			vs.SetClass(panelH, classname)
	
			return panelH
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "Pyramid Panel" )
		this.SetListAbsorb()			 #  mark this node as list absorbing
		this.SetDescription( 'Extrude pyramidal definition.' )
	
		#Input Ports
		shapes = Marionette.PortIn(vs.Handle(0), 'ShapeH(s)')
		shapes.SetDescription( "Shapes that define single panel size" )
		
		classname = Marionette.PortIn(0, 'sClass')
		classname.SetDescription( "Class name" )
	
		solid = Marionette.PortIn(False, 'bSolid')
		solid.SetDescription( "Make panel solid" )
	
		bulge = Marionette.PortIn(0, 'nBulge')
		bulge.SetDescription( "Panel bulge at peak" )
	
		edgeoffset = Marionette.PortIn(0, 'nEdge Offset')
		edgeoffset.SetDescription( "Panel edge offset" )
		
		hgtoffset = Marionette.PortIn(0, 'nHgt Offset')
		hgtoffset.SetDescription( "Panel height offset" )
	
		bezelangle = Marionette.PortIn(0, 'nBezel Angle')
		bezelangle.SetDescription( "Panel edge bezel angle" )
		
		thickness = Marionette.PortIn(0, 'nAvg Thickness')
		thickness.SetDescription( "Panel average thickness" )
	
		gasket = Marionette.PortIn(0, 'Gasketdef')
		gasket.SetDescription( "Panel gasket definition" )
	
		flipsides = Marionette.PortIn(False, 'bFlip Sides')
		flipsides.SetDescription( "Flip which side the panel is drawn" )
	
		#Output Ports
		panels = Marionette.PortOut('panelH(s)')
		panels.SetDescription( "The resultant panel objects" )
	
		status = Marionette.PortOut('sStatus')
		status.SetDescription( "The status of the created Panel object(s)" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
		def getboolval(bvalue):
			if isinstance(bvalue,list):
				return bvalue[0]
			else:
				return bvalue
	
		def getnumbval(value,defvalue):
			if isinstance(value,(list,tuple)):
				value = value[0]
			if isinstance(value,(int,float)):
				return value
			else:
				return defvalue
	
		#inputs
		shapes = self.Params.getinputObject(self.Params.shapes.value)
		status = 0
		
		if shapes:
			classname   = self.Params.getclass(self.Params.classname.value)
			
			gasketdef = self.Params.gasket.value
			if isinstance(gasketdef,list) and ('PNG' in gasketdef[0]):
				gasket = True
				gasketdef = gasketdef[0]['PNG']
			else:
				gasket = False
			
			bulge	   = getnumbval(self.Params.bulge.value,0)
			edgeoffset = getnumbval(self.Params.edgeoffset.value,0)
			hgtoffset  = getnumbval(self.Params.hgtoffset.value,0)
			bezelangle = getnumbval(self.Params.bezelangle.value,0)
			thickness  = getnumbval(self.Params.thickness.value,0)
			solid	   = getboolval(self.Params.solid.value)
			flipsides  = getboolval(self.Params.flipsides.value)
	
			paneldef = {'panelshape' : 'Pyramidal',
			'solid' : str(bool(solid)),
			'bulge' : str(bulge),
			'edgeoffset' : str(edgeoffset),
			'hgtoffset'  : str(hgtoffset),
			'bezelangle' : str(bezelangle),
			'thickness'  : str(thickness),
			'flipsides'  : str(bool(flipsides)),
			'withgaskets' : str(gasket)
			}
			
			if gasket:
				paneldef.update(gasketdef)
	
			if len(shapes) == 1: # single shape
				panelpios = self.Params.drawpanel(shapes[0],classname,paneldef)
			else:
				panelpios = []
				status = []
				vs.BeginGroup()
				for shape in shapes: # list of shapes
					panelH = self.Params.drawpanel(shape,classname,paneldef)
					panelpios.append(panelH)
					objstatus = vs.GetRField(panelH,'Extrude Panel Planar','status')
					if objstatus != '':
						status.append(objstatus)
				vs.EndGroup()
				panelpios = (vs.LNewObj(),panelpios)
		else:
			panelpios = 0 # fail
			status = 'No valid path shapes'
	
		#outputs
		self.Params.panels.value = panelpios
		self.Params.status.value = status
	