	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Place/Get 3D Poly Definitions.py;
	# Created by Bill Wood 25/4/20
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if (type(inputObject) == vs.Handle): # single object
				objlistH = [inputObject]
			else:
				objlistH = inputObject # list of objects??
			return objlistH
			
			newobjlistH = []
			if isinstance(objlistH,list):
				for objH in objlistH:
					objtype = vs.GetTypeN(objH)
					if objtype == 25: # 3D poly
						newobjlistH.append(objH)
				if len(newobjlistH) == 0:
					return 0 # invalid objects in objectlist
				else:
					return newobjlistH
			else:
				return 0 # not a list
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "Get 3D Poly Definitions" )
		this.SetListAbsorb()			 #  mark this node as list absorbing
		this.SetDescription( "Get definition of a supplied group of 3D polygon(s)." )
		
		#Input Ports
		polys = Marionette.PortIn(vs.Handle(0), '3D poly(s)')
		polys.SetDescription( "3D poly(s) to get definition from" )
	
		#Output Ports
		polydef = Marionette.PortOut('PolyDef(s)')
		polydef.SetDescription( "3D polygon(s) Face centroid and orientation" )
	
			
	def RunNode(self):
	
		def getmatrix(objH,pt1,pt2):
			btest,x1,y1,z1 = vs.Centroid3D(objH)
			pt3 = (x1,y1,z1)
	
			v1 = (pt2[0]-pt1[0],pt2[1]-pt1[1],pt2[2]-pt1[2])
			v2 = (pt3[0]-pt1[0],pt3[1]-pt1[1],pt3[2]-pt1[2])
			v3 = vs.CrossProduct(v1, v2)
			v2 = vs.CrossProduct(v1, v3)
			v2 = (-v2[0],-v2[1],-v2[2])
	
			mati = vs.UnitVec(v1)
			matj = vs.UnitVec(v2)
			matk = vs.UnitVec(v3)
	
			return(mati,matj,matk,(pt3[0],pt3[1],pt3[2]))
			# getmatrix
	
		def addvertex(vert,fcnt):
			nonlocal verts
			nonlocal vertfaces
	
			def vertinverts(refvert,verts):
				vref = []
				for i in range(len(verts)):
					v1 = (refvert[0],refvert[1],refvert[2])
					v2 = (verts[i][0],verts[i][1],verts[i][2])
					if vs.EqPt3D(v1,v2,0.00001): # used for rounding
						vref = i
						break
				return vref
	
			vref = vertinverts(vert,verts)
			if vref != []:
				if fcnt not in vertfaces[vref]:
					vertfaces[vref].append(fcnt)
			else:
				verts.append(vert) # add to vertex list
				fref = [fcnt]
				vertfaces.append(fref)
				vref = len(verts)-1 # first in vertex list
	
			return vref
			# addvertex
	
		def setedgevec(edge):
			nonlocal verts
			nonlocal edgevecs
	
			pt1 = verts[edge[0]]
			pt2 = verts[edge[1]]
			
			vec = (pt2[0]-pt1[0],pt2[1]-pt1[1],pt2[2]-pt1[2])
			midpos = (pt1[0]+vec[0]/2,pt1[1]+vec[1]/2,pt1[2]+vec[2]/2)
			edgevec = vs.UnitVec(vec)
	
			# store edge mid position and edge delta X axis direction
			edgevecs.append((midpos,edgevec))
	
			return
			# setedgevec
				
		def addedge(edge,fcnt):
			nonlocal edges
			nonlocal edgefaces
	
			if edge[0] > edge[1]: # order low to high
				edge = (edge[1],edge[0])
	
			if edge in edges:
				edgeref = edges.index(edge)
				if fcnt not in edgefaces[edgeref]:
					edgefaces[edgeref].append(fcnt)
				else:
					fdef = [fcnt]
					edgefaces[edgeref] = fdef # new entry
			else:
				edges.append(edge) # add to edge list
				setedgevec(edge)
	
				fdef = [fcnt]
				edgefaces.append(fdef)
	
			return
			# addedge
	
		def getfacedef(faceH,fcnt):
			nonlocal faces
	
			pt1 = vs.GetPolyPt3D(faceH,0)
			pt2 = vs.GetPolyPt3D(faceH,1)
			xaxis,yaxis,zaxis,centpos = getmatrix(faceH,pt1,pt2)
	
			vlocfrom = addvertex(pt1,fcnt)
			startfrom = vlocfrom
			vlocto   = addvertex(pt2,fcnt)
			
			addedge((vlocfrom,vlocto),fcnt)
			faces.append((xaxis,yaxis,zaxis,centpos))
			vlocfrom = vlocto
	
			vcnt = vs.GetVertNum(faceH)
			for i in range(2,vcnt):
				vert = vs.GetPolyPt3D(faceH, i)
				vlocto   = addvertex(vert,fcnt)
				addedge((vlocfrom,vlocto),fcnt)
				vlocfrom = vlocto
	
			addedge((vlocfrom,startfrom),fcnt) # close polygon
	
			return
			#getfacedef
	
		# RunNode
		polys = self.Params.getinputObject(self.Params.polys.value)
	
		polynum = len(polys)
		if polynum > 0:
			faces = [] 		# face entry 	= (xaxis,yaxis,zaxis,centpos)
			edges = [] 		# edge entry 	= (fromvertno,tovertno)
			edgevecs  = [] 	# edge entry 	= (midpos,edgevec)
			edgefaces = [] 	# edge entry 	= (faceno1,faceno2,..faceN)
			verts = [] 		# vertex entry 	= (vertexpos)
			vertfaces = []	# vertex entry	= (faceno1,faceno2,..faceN)
	
			for i in range(polynum):
				getfacedef(polys[i],i)
				polydef = {'PD':{'faces':faces,'vertices':verts,'edges':edges,'edgevecs':edgevecs,'edgefaces':edgefaces,'vertfaces':vertfaces}}
		else:
			polydef = 0
	
		self.Params.polydef.value = polydef