	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Place/Linear Place on edges UI.py;
	# Created by Bill Wood 25/4/20
	# Edited
	# Renamed from Place on Edges UI. Vertical option fixed 24/2/21
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputPolydefs(inputObject):
			polydefs = 0
			if isinstance(inputObject,list):
				if len(inputObject) == 1: # Only first polydef logged Multiple PD not allowed
					pddef = inputObject[0]
					polydefs = pddef['PD']
			return polydefs
	
		def getinputShape(inputShape):
			result = {'shapedef':'Standard','shapetype':'Rectangle','__standardshape':'1', 'val1':'10mm','val2':'10mm','thickness':'0'} # Default
			if isinstance(inputShape,list):
				inputShape = inputShape[0] # only one entry allowed - others ignored
	
			if isinstance(inputShape,dict):
				if 'IS' in inputShape:
					result = inputShape['IS']
					
			return result
	
		def getorientation(orientation):
			result = {'xoffset':'0','yoffset':'0','shapeangle':'0','fliphori':'False','flipvert':'False','shapepivot':'As Shape'} # default
			if isinstance(orientation,list):
			 	orientation = orientation[0] # only one entry allowed - others ignored
	
			if isinstance(orientation,dict):
				if 'SO' in orientation:
					result = orientation['SO']
	
			return result
	
		def getmitres(mitres):
			result = {'startmitre':'None','endmitre':'None'} # Default
			if isinstance(mitres,list):
			 	mitres = mitres[0] # only one entry allowed - others ignored
	
			if isinstance(mitres,dict):
				if 'MI' in mitres:
					result = mitres['MI']
	
			return result
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "EP Linear on Edges UI" )
		this.SetListAbsorb()			 #  mark this node as list absorbing
		this.SetDescription( "Create EP Extrude Linear path plugin object(s) based on supplied edge definition. Note edges with linked faces will create 2 Linear path objects" )
	
		#Input Ports
		polydefs = Marionette.PortIn(0, 'PolyDef(s)' )
		polydefs.SetDescription( "Definition of poly faces/edges paths for the EP Extrude Linear object(s)" )
	
		inputShape = Marionette.PortIn(0, 'Shape' )
		inputShape.SetDescription( "The input shape definition for the EP Extrude Linear object" )
	
		orientation = Marionette.PortIn(0, 'Orientation' )
		orientation.SetDescription( "The shape orientation for the EP Extrude Linear object" )
	
		mitredef = Marionette.PortIn(0, 'Mitres' )
		mitredef.SetDescription( "The end mitre definitions for the EP Linear object" )
	
		#OIP Input Controls
		classname = Marionette.OIPControl( 'Class', Marionette.WidgetType.PopupClasses, 'None')	
		classname.SetDescription( "The shape classname used for the EP Extrude Linear object" )
	
		startoffset = Marionette.OIPControl('Start Offset', Marionette.WidgetType.RealCoord, 0)
		startoffset.SetDescription( "Path offset at start" )
	
		endoffset = Marionette.OIPControl('End Offset', Marionette.WidgetType.RealCoord, 0)
		endoffset.SetDescription( "Path offset at end" )
		
		bulge = Marionette.OIPControl('Curve Bulge', Marionette.WidgetType.RealCoord, 0)
		bulge.SetDescription( "Set EP Linear plugin with curve bulge distance" )
	
		reversepath = Marionette.OIPControl('Reverse Path', Marionette.WidgetType.Bool, False)
		reversepath.SetDescription( "Reverse direction of path object" )
	
		showpath = Marionette.OIPControl('Show Path', Marionette.WidgetType.Bool, False)
		showpath.SetDescription( "Show path object route" )
	
		vertical = Marionette.OIPControl( 'Force Vertical', Marionette.WidgetType.Bool, False)
		vertical.SetDescription( "Orient Linear shapes vertical" )
	
		#Output Ports
		extrH = Marionette.PortOut('hExtrude(s)')
		extrH.SetDescription( "The result EP Extrude Linear object or grouped Extrude Linear objects" )
	
		status = Marionette.PortOut('sStatus')
		status.SetDescription( "The status of the created EP Extrude Linear object(s)" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
		def addvec(dir,axis):
			return (dir[0] + axis[0],dir[1] + axis[1],dir[2] + axis[2])
				
		def drawextrudebyline(startpos,linelen,xdir,ydir,zdir,params,classname):
			extrudetype = 'Extrude Linear'
			extrH = vs.CreateCustomObjectN(extrudetype,0,0,0,False)
	
			if vs.SetEntityMatrixN(extrH,xdir,ydir,zdir,startpos):
				vs.SetRField(extrH,extrudetype,'LineLength',str(linelen))
				for k in params:
					vs.SetRField(extrH,extrudetype,k,params[k])
				vs.SetClass(extrH, classname)
				status = vs.GetRField(extrH,extrudetype,'status')
			else:
				status = 'Error in setting EP Linear orientation'
				vs.DelObject(extrH)
				extrH = 0;
	
			return (status,extrH)
	
		#Inputs
		polydefs	= self.Params.getinputPolydefs(self.Params.polydefs.value)
		status = 0
		if polydefs:
			#Other inputs
			shapeData   = self.Params.getinputShape(self.Params.inputShape.value)
			orientation = self.Params.getorientation(self.Params.orientation.value)
			mitres		= self.Params.getmitres(self.Params.mitredef.value)
	
			#UI controls
			classname   = self.Params.classname.value
			startoffset = self.Params.startoffset.value
			endoffset   = self.Params.endoffset.value
			bulge	   = self.Params.bulge.value
			reversepath = self.Params.reversepath.value
			showpath	= self.Params.showpath.value
			vertical	= self.Params.vertical.value
	
			faces	  = polydefs['faces']
			vertices  = polydefs['vertices']
			edges	  = polydefs['edges']
			edgevecs  = polydefs['edgevecs']
			edgefaces = polydefs['edgefaces']
			vertfaces = polydefs['vertfaces']
	
			params = shapeData
			params.update(orientation)
			params.update(mitres)
			options = {'startoffset':str(startoffset),'endoffset':str(endoffset),'reversepath':str(bool(reversepath)),'showloci':str(bool(showpath)),'bulge':str(bulge)}
			params.update(options)
	
			extrHpios = []
			for i in range(len(edgevecs)):
				edge = edges[i]
				pos1 = vertices[edge[0]]
				pos2 = vertices[edge[1]]
				linelen = vs.Distance3D(pos1[0],pos1[1],pos1[2],pos2[0],pos2[1],pos2[2])
	
				edgevec = edgevecs[i]
				xdir	= edgevec[1] # always returns unit vector
					
				facedef = edgefaces[i]
				if vertical: # and not ((xdir == (0,0,1)) or (xdir == (0,0,-1)))
					for i in range(2):
						faceno = facedef[i]
						if i == 1:
							face1 = facedef[0]
							face2 = facedef[1]
							face1Z = faces[face1][2]
							face2Z = faces[face2][2]
							facetest = not vs.EqPt3D(face1Z,face2Z,0.00001) # used for rounding
						else:
							facetest = True
						
						if facetest:
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = vs.UnitVec(zaxis)
							ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
								
							piostatus,pioH = drawextrudebyline(pos1,linelen,xdir,ydir,zdir,params,classname)
							if pioH:
								extrHpios.append(pioH)
							else:
								status = piostatus
				else:
					zdir = (0,0,0)
					for faceno in facedef:
						xaxis,yaxis,zaxis,centpos = faces[faceno]
						zdir = addvec(zdir,zaxis)
	
					zdir = vs.UnitVec(zdir)
					ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
	
					piostatus,pioH = drawextrudebyline(pos1,linelen,xdir,ydir,zdir,params,classname)
					if pioH:
						extrHpios.append(pioH)
					else:
						status = piostatus
				
		else:
			extrHpios = 0
			status = 'No valid poly definitions'
				
		#outputs
		self.Params.extrH.value  = extrHpios
		self.Params.status.value = status