	@Marionette.NodeDefinition
	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Place/Symbol Place UI.py;
	# Created by Bill Wood 25/4/20
	
	class Params(metaclass = Marionette.OrderedClass):
		def getinputSymboldefs(inputObject):
			# inputObject always return a list if content passed
			symlist = []
			def setsymname(symname):
				nonlocal symlist
				name = symname.lower()
				if symlist.count(name) == 0:
					symlist.append(name)
			
			def checksymname(symdef): # symbol definition
				symH = vs.GetObject(symdef)
				return (vs.GetTypeN(symH) == 16)
	
			if isinstance(inputObject,list): # Always supplied as a list because of SetListAbsorb
				for symdef in inputObject:
					if (type(symdef) == vs.Handle): # single object
						# check for symbol instance or symbol definition
						objty = vs.GetTypeN(symdef)
						if  objty == 15: # symbol instance
							setsymname(vs.GetSymName(symdef))
						elif objty == 16: # symbol definition
							setsymname(vs.GetSDName(symdef))
	
					elif isinstance(symdef,str) and checksymname(symdef):
						setsymname(symdef)
	
				if len(symlist) == 0:
					symlist = 0
			else:
				symlist = 0
			return symlist
	
		def getinputPolydefs(inputObject):
			polydefs = 0
			if isinstance(inputObject,list):
				if len(inputObject) == 1: # Only first polydef logged Multiple PD not allowed
					pddef = inputObject[0]
					polydefs = pddef['PD']
			return polydefs
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "EP Symbol Place UI" )
		this.SetListAbsorb()			 #  mark this node as list absorbing
		this.SetDescription( "Place symbol(s) in relation to polygon definitions" )
	
		#Input Ports
		polydefs = Marionette.PortIn([], 'PolyDef(s)')
		polydefs.SetDescription( "Definition of poly faces" )
	
		symboldefs = Marionette.PortIn(vs.Handle(0), 'Symbol(s)')
		symboldefs.SetDescription( "Symbol names or symbol definition/instance Handles" )
	
		#OIP Input Controls
		atvertices = Marionette.OIPControl( 'At Vertices', Marionette.WidgetType.Bool, True)
		atvertices.SetDescription( "Place objects at polygon vertices" )
	
		atmidedges = Marionette.OIPControl( 'At Mid Edges', Marionette.WidgetType.Bool, False)
		atmidedges.SetDescription( "Place objects at polygon mid edges" )
	
		atedgeends = Marionette.OIPControl( 'At Edge Ends', Marionette.WidgetType.Bool, False)
		atedgeends.SetDescription( "Place objects at edge ends aligned with edge" )
	
		atfacecentres = Marionette.OIPControl( 'At Face Centres', Marionette.WidgetType.Bool, False)
		atfacecentres.SetDescription( "Place objects at polygon face centres" )
	
		classname = Marionette.OIPControl( 'Class', Marionette.WidgetType.PopupClasses, 'None')
		classname.SetDescription( "List of classes" )
	
		texttitle = Marionette.OIPControl( 'Offset', Marionette.WidgetType.TextStatic, '')
	
		xoffset = Marionette.OIPControl( '-  in X', Marionette.WidgetType.RealCoord, 0.0)
		xoffset.SetDescription('Object offset along symbol X axis')
		
		yoffset = Marionette.OIPControl( '-  in Y', Marionette.WidgetType.RealCoord, 0.0)
		yoffset.SetDescription('Object offset along symbol Y axis')
	
		zoffset = Marionette.OIPControl( '-  in Z', Marionette.WidgetType.RealCoord, 0.0)
		zoffset.SetDescription('Object offset along symbol Z axis')
	
		vertical = Marionette.OIPControl( 'Always Vertical', Marionette.WidgetType.Bool, False)
		vertical.SetDescription( "Placed objects always vertical" )
	
		flipdirection = Marionette.OIPControl( 'Flip Direction', Marionette.WidgetType.Bool, False)
		flipdirection.SetDescription( "Placed objects flipped about XY layer plane" )
	
		#Output Ports
		symbols = Marionette.PortOut('hSymbol(s)')
		symbols.SetDescription( "The resultant placed symbols(s)" )
	
		status = Marionette.PortOut('sStatus')
		status.SetDescription( "The status of the created symbols(s)" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
	
		def placesymbols(sym,offset,classname,atvertices,atmidedges,atedgeends,atfacecentres,vertical,flipdirection,faces,vertices,edges,edgevecs,edgefaces,vertfaces):
	
			def createsym(sym,classname,locdef,offset):
				nonlocal symbolpios
				axisX,axisY,axisZ,origin = locdef
				xoffset = offset[0]
				yoffset = offset[1]
				zoffset = offset[2]
	
				if (xoffset != 0):
					origin = (origin[0]+axisX[0]*xoffset,origin[1]+axisX[1]*xoffset,origin[2]+axisX[2]*xoffset)
				if (yoffset != 0):
					origin = (origin[0]+axisY[0]*yoffset,origin[1]+axisY[1]*yoffset,origin[2]+axisY[2]*yoffset)
				if (zoffset != 0):
					origin = (origin[0]+axisZ[0]*zoffset,origin[1]+axisZ[1]*zoffset,origin[2]+axisZ[2]*zoffset)
	
				vs.Symbol(sym,(0,0),0)
				symH = vs.LNewObj()
				if vs.SetEntityMatrixN(symH, axisX, axisY, axisZ, origin):
					vs.SetClass(symH,classname) 
					symbolpios.append(symH)
				else:
					vs.DelObject(symH)
				return
			
			def addvec(dir,axis):
				return (dir[0] + axis[0],dir[1] + axis[1],dir[2] + axis[2])
			
			def flipvec(vec):
				return (-vec[0],-vec[1],-vec[2])
			
			def setatfacecentres():
				for face in faces:
					if vertical:
						xdir = (1,0,0)
						ydir = (0,1,0)
						zdir = (0,0,-1) if flipdirection else (0,0,1)
						origin = face[3]
					else:
						xdir,ydir,zdir,origin = face
	
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
					createsym(sym,classname,(xdir,ydir,zdir,origin),offset)
				return
				
			def setatvertices():
				if vertical:
					# reset X, Y and Z axes if used
					xdir = (1,0,0); ydir = (0,1,0)
					zdir = (0,0,-1) if flipdirection else (0,0,1)
	
				for vno in range(len(vertices)):
					vertexpos = vertices[vno]
	
					if not vertical:
						xdir = (0,0,0)
						ydir = (0,0,0)
						zdir = (0,0,0)
	
						facenos = vertfaces[vno]
						for j in range(len(facenos)):
							faceno = facenos[j]
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = addvec(zdir,zaxis)
	
						zdir = vs.UnitVec(zdir)
						if vs.EqPt3D((zdir[0],zdir[1],abs(zdir[2])),(0,0,1),0.001): # zdir is vertical
							xdir = (1,0,0) if zdir[2] > 0 else (-1,0,0)
						else:
							xdir = vs.Perp(zdir)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
					createsym(sym,classname,(xdir, ydir, zdir, vertexpos),offset)
				return
			
			def setatmidedges():
				for i in range(len(edgefaces)):
					edgevec = edgevecs[i]
					midpos = edgevec[0]
					
					if vertical:
						# reset X, Y and Z axes
						xdir = (1,0,0); ydir = (0,1,0)
						zdir = (0,0,-1) if flipdirection else (0,0,1)
					else:
						facedef = edgefaces[i]
						xdir	= edgevec[1]
						zdir = (0,0,0)
						for faceno in facedef:
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = addvec(zdir,zaxis)
	
						zdir = vs.UnitVec(zdir)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
	
					createsym(sym,classname,(xdir, ydir, zdir, midpos),offset)
				return
			
			def setatedgeends():
				for i in range(len(edgefaces)):
					edgevec = edgevecs[i]
					startpos = vertices[edges[i][0]]
					endpos   = vertices[edges[i][1]]
	
					if vertical:
						# reset X, Y and Z axes
						xdir = (1,0,0); ydir = (0,1,0)
						zdir = (0,0,-1) if flipdirection else (0,0,1)
						
						#check
					else:
						facedef = edgefaces[i]
	
						zdir	= edgevec[1]
						ydir = (0,0,0)
						for faceno in facedef:
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							ydir = addvec(ydir,zaxis)
	
						ydir = vs.UnitVec(ydir)
						xdir = vs.CrossProduct(zdir,ydir) # always returns unit vector?
	
						if flipdirection:
							xdir = flipvec(xdir)
							ydir = flipvec(ydir)
						#else:
						#	zdir = flipvec(zdir)
	
					createsym(sym,classname,(xdir, ydir, zdir, startpos),offset)
					#xdir = flipvec(xdir)
					#ydir = flipvec(ydir)
					if not vertical:
						zdir = flipvec(zdir)
						createsym(sym,classname,(xdir, ydir, zdir, endpos),offset)
	
				return
	
			# ===== placesymbols
			if atfacecentres:
				setatfacecentres()
			
			if atvertices:
				setatvertices()
				
			if atmidedges:
				setatmidedges()
	
			if atedgeends:
				setatedgeends()
	
			return
	
		# ===== RunNode
		status = 0
		symbolpios = []
	
		# Node controls
		symbols  = self.Params.getinputSymboldefs(self.Params.symboldefs.value)
		polydefs = self.Params.getinputPolydefs(self.Params.polydefs.value)
	
		# UI controls
		atvertices = self.Params.atvertices.value
		atmidedges = self.Params.atmidedges.value
		atedgeends = self.Params.atedgeends.value
		atfacecentres = self.Params.atfacecentres.value
		atset = (atvertices or atmidedges or atedgeends or atfacecentres)
	
		if (symbols and polydefs and atset):
			classname	  = self.Params.classname.value
			offset = (self.Params.xoffset.value,self.Params.yoffset.value,self.Params.zoffset.value)
			vertical	  = self.Params.vertical.value
			flipdirection = self.Params.flipdirection.value
			
			faces	  = polydefs['faces']
			vertices  = polydefs['vertices']
			edges	  = polydefs['edges']
			edgevecs  = polydefs['edgevecs']
			edgefaces = polydefs['edgefaces']
			vertfaces = polydefs['vertfaces']
	
			for sym in symbols:
				placesymbols(sym,offset,classname,atvertices,atmidedges,atedgeends,atfacecentres,vertical,flipdirection,faces,vertices,edges,edgevecs,edgefaces,vertfaces)
	
		else:
			if not polydefs:
				status = 'Invalid poly definition(s) or Multi poly definition'
			elif not symbols:
				status = 'Invalid symbol definition(s)'
			elif not atset:
				status = 'No place at settings'
	
		#outputs
		self.Params.symbols.value = symbolpios
		self.Params.status.value = status
	