	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Place/Place Shape UI.py;
	# Created by Bill Wood 25/4/20
	
	@Marionette.NodeDefinition
	class Params(metaclass = Marionette.OrderedClass):
		def getinputObject(inputObject):
			if isinstance(inputObject,list):
				inputObject = inputObject[0] # only one entry allowed - others ignored
			result = 0 # default fail
			if isinstance(inputObject,dict) and ('EI' in inputObject):
				# {'EI':[location,{'topoffset':,'extrudehgt':,'baseZ':,'skewX':,'skewY':,'verticaloffsets':,'maintainshape':}]}
					content = inputObject['EI']
					# location ignored
					result = content[1] # params
			else: # default values
				result = 0 # default fail
		
			return result
	
		def getinputShape(inputShape):
			result = {'shapedef':'Standard','shapetype':'Rectangle','__standardshape':'1', 'val1':'10mm','val2':'10mm','thickness':'0'} # Default
			if isinstance(inputShape,list):
				inputShape = inputShape[0] # only one entry allowed - others ignored
	
			if isinstance(inputShape,dict):
				if 'IS' in inputShape:
					result = inputShape['IS']
					
			return result
	
		def getorientation(orientation):
			result = {'xoffset':'0','yoffset':'0','shapeangle':'0','fliphori':'False','flipvert':'False','shapepivot':'As Shape'} # default
			if isinstance(orientation,list):
			 	orientation = orientation[0] # only one entry allowed - others ignored
	
			if isinstance(orientation,dict):
				if 'SO' in orientation:
					result = orientation['SO']
	
			return result
	
		def getinputPolydefs(inputObject):
			polydefs = 0
			if isinstance(inputObject,list):
				if len(inputObject) == 1: # Only first polydef logged Multiple PD not allowed
					pddef = inputObject[0]
					polydefs = pddef['PD']
			return polydefs
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "EP Point Place UI" )
		this.SetListAbsorb()			 #  mark this node as list absorbing
		this.SetDescription( 'Place EP Point(s) ojbects in relation to polygon definitions' )
	
		#Input Ports
		polydefs = Marionette.PortIn([], 'PolyDef(s)')
		polydefs.SetDescription( "Definition of poly faces" )
		
		inputOptions = Marionette.PortIn(vs.Handle(0), 'Options')
		inputOptions.SetDescription( "The input options for the EP Shape object" )
	
		inputShape = Marionette.PortIn(0, 'Shape' )
		inputShape.SetDescription("EP Shape type")
	
		orientation = Marionette.PortIn(0, 'Orientation' )
		orientation.SetDescription( "The shape orientation for the EP Shape object" )
	
		atvertices = Marionette.OIPControl( 'At Vertices', Marionette.WidgetType.Bool, True)
		atvertices.SetDescription( "Place objects at polygon vertices" )
	
		atmidedges = Marionette.OIPControl( 'At Mid Edges', Marionette.WidgetType.Bool, False)
		atmidedges.SetDescription( "Place objects at polygon mid edges" )
	
		atfacecentres = Marionette.OIPControl( 'At Face Centres', Marionette.WidgetType.Bool, False)
		atfacecentres.SetDescription( "Place objects at polygon face centres" )
	
		classname = Marionette.OIPControl( 'Class', Marionette.WidgetType.PopupClasses, 'None')
		classname.SetDescription( "List of classes" )
	
		vertical = Marionette.OIPControl( 'Always Vertical', Marionette.WidgetType.Bool, False)
		vertical.SetDescription( "Placed objects always vertical" )
	
		# Set in extrDef for linked node "EP Extrude Shape Input"
		# zoffset = nBottomZ also nTopOffset can be set
		# flipdirection = nHeight set as minus 
	
		#Output Ports
		extrH = Marionette.PortOut('hExtrude(s)')
		extrH.SetDescription( "The resultant EP Shape object or grouped Shape objects" )
	
		status = Marionette.PortOut('sStatus')
		status.SetDescription( "The status of the created EP Shape object(s)" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
	
		def placeobjects(params,classname,atvertices,atmidedges,atfacecentres,vertical,faces,vertices,edges,edgevecs,edgefaces,vertfaces):
			nonlocal extrpios
	
			def createpio(params,classname,locdef):
				nonlocal extrpios
				extrudetype = 'Extrude Point'
				
				axisX,axisY,axisZ,origin = locdef
	
				extrH = vs.CreateCustomObject(extrudetype,0,0,0)
				if vs.SetEntityMatrixN(extrH,axisX,axisY,axisZ,origin):
	
					for k in params:
						vs.SetRField(extrH,extrudetype,k,params[k])
					vs.SetClass(extrH, classname)
					status = vs.GetRField(extrH,extrudetype,'status')
	
					extrpios.append(extrH)
				else:
					status = 'Error in setting EP Point orientation'
					vs.DelObject(extrH)
					extrH = 0;
	
				return (status,extrH)
			
			def addvec(dir,axis):
				return (dir[0] + axis[0],dir[1] + axis[1],dir[2] + axis[2])
			
			def flipvec(vec):
				return (-vec[0],-vec[1],-vec[2])
	
			# ===== placeobjects
	
			if atfacecentres:
				for face in faces:
					if vertical:
						xdir = (1,0,0); ydir = (0,1,0); zdir = (0,0,1)
						origin = face[3]
					else:
						xdir,ydir,zdir,origin = face
						ydir = flipvec(ydir)
						zdir = flipvec(zdir)
	
					createpio(params,classname,(xdir,ydir,zdir,origin))
			
			if atvertices:
				if vertical:
					# reset X, Y and Z axes if used
					xdir = (1,0,0); ydir = (0,1,0); zdir = (0,0,1)
	
				for vno in range(len(vertices)):
					vertexpos = vertices[vno]
	
					if not vertical:
						xdir = (0,0,0)
						ydir = (0,0,0)
						zdir = (0,0,0)
	
						facenos = vertfaces[vno]
						for j in range(len(facenos)):
							faceno = facenos[j]
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = addvec(zdir,zaxis)
						
						zdir = vs.UnitVec(zdir)
						if vs.EqPt3D((zdir[0],zdir[1],abs(zdir[2])),(0,0,1),0.001): # zdir is vertical
							xdir = (1,0,0) if zdir[2] > 0 else (-1,0,0)
						else:
							xdir = vs.Perp(zdir)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						ydir = flipvec(ydir)
						zdir = flipvec(zdir)
	
					createpio(params,classname,(xdir,ydir,zdir,vertexpos))
				
			if atmidedges:
				for i in range(len(edgefaces)):
					edgevec = edgevecs[i]
					midpos = edgevec[0]
					
					if vertical:
						# reset X, Y and Z axes
						xdir = (1,0,0); ydir = (0,1,0); zdir = (0,0,1)
					else:
						facedef = edgefaces[i]
						xdir	= edgevec[1]
						zdir = (0,0,0)
						for faceno in facedef:
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = addvec(zdir,zaxis)
	
						zdir = vs.UnitVec(zdir)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						ydir = flipvec(ydir)
						zdir = flipvec(zdir)
						
					createpio(params,classname,(xdir,ydir,zdir,midpos))
			return
	
		status = 0
		extrpios = []
	
		# Node controls
		inputOptions = self.Params.getinputObject(self.Params.inputOptions.value)
		polydefs	 = self.Params.getinputPolydefs(self.Params.polydefs.value)
			
		# UI controls
		atvertices = self.Params.atvertices.value
		atmidedges = self.Params.atmidedges.value
		atfacecentres = self.Params.atfacecentres.value
		atset = atvertices or atmidedges or atfacecentres
	
		if inputOptions and polydefs and atset:
			shapeData   = self.Params.getinputShape(self.Params.inputShape.value)
			orientation = self.Params.getorientation(self.Params.orientation.value)
		
			classname  = self.Params.classname.value
			vertical   = self.Params.vertical.value		
	
			faces	  = polydefs['faces']
			vertices  = polydefs['vertices']
			edges	  = polydefs['edges']
			edgevecs  = polydefs['edgevecs']
			edgefaces = polydefs['edgefaces']
			vertfaces = polydefs['vertfaces']
	
			params = shapeData
			params.update(orientation)
			params.update(inputOptions)
	
			placeobjects(params,classname,atvertices,atmidedges,atfacecentres,vertical,faces,vertices,edges,edgevecs,edgefaces,vertfaces)
	
		else:
			if not polydefs:
				status = 'Invalid poly definition(s)'
			elif not inputOptions:
				status = 'Invalid object definition(s)'
			elif not atset:
				status = 'No place at settings'
			
	
		#outputs
		self.Params.extrH.value  = extrpios
		self.Params.status.value = status
	