	#COMMAND;REFFILE;[UsrLib]/Defaults/Marionette/WW Tools/Place/Place Symbol.py;
	# Created by Bill Wood 25/4/20
	# Edited
	# Title renamed from EP Place Symbol 24/2/21
	
	@Marionette.NodeDefinition
	class Params(metaclass = Marionette.OrderedClass):
		def getinputSymboldefs(inputObject):
			# inputObject always return a list if content passed
			symlist = []
			def setsymname(symname):
				nonlocal symlist
				name = symname.lower()
				if symlist.count(name) == 0:
					symlist.append(name)
			
			def checksymname(symdef): # symbol definition
				symH = vs.GetObject(symdef)
				return (vs.GetTypeN(symH) == 16)
	
			if not isinstance(inputObject,list):
				inputObject = list(inputObject)
	
			for symdef in inputObject:
				if (type(symdef) == vs.Handle): # single object
					# check for symbol instance or symbol definition
					objty = vs.GetTypeN(symdef)
					if  objty == 15: # symbol instance
						setsymname(vs.GetSymName(symdef))
					elif objty == 16: # symbol definition
						setsymname(vs.GetSDName(symdef))
	
				elif isinstance(symdef,str) and checksymname(symdef):
					setsymname(symdef)
	
			if len(symlist) == 0:
				symlist = 0
	
			return symlist
	
		def getinputPolydefs(inputObject):
			polydefs = 0
			if isinstance(inputObject,list):
				if len(inputObject) == 1: # Only first polydef logged Multiple PD not allowed
					pddef = inputObject[0]
					polydefs = pddef['PD']
			return polydefs
	
		def getclass(classname):
			clname = 'None' # default class name
			if isinstance(classname,str):
				clindex = vs.Name2Index(classname)
				if vs.ClassList(clindex-1) == classname: # check class exists
					clname = classname
			return clname
	
		def getboolval(bvalue):
			if isinstance(bvalue,list):
				return bvalue[0]
			else:
				return bvalue
	
		def getnumbval(value,defvalue):
			if isinstance(value,(list,tuple)):
				value = value[0]
			if isinstance(value,(int,float)):
				return value
			else:
				return defvalue
	
	#APPEARANCE
		#Name
		this = Marionette.Node( "EP Symbol Place" )
		this.SetListAbsorb()			 #  mark this node as list absorbing
		this.SetDescription( "Place symbol(s) in relation to polygon definitions" )
	
		#Input Ports
		polydefs = Marionette.PortIn([], 'PolyDef(s)')
		polydefs.SetDescription( "Definition of poly faces" )
	
		symboldefs = Marionette.PortIn(vs.Handle(0), 'Symbol(s)')
		symboldefs.SetDescription( "Symbol names or instance Handles" )
		
		# Note must set atvertices,atmidedges and or atfacecentres
		atvertices = Marionette.PortIn(False, 'bVertices')
		atvertices.SetDescription( "Place objects at poly vertices" )
	
		atmidedges = Marionette.PortIn(False, 'bMidEdges')
		atmidedges.SetDescription( "Place objects at poly mid edges" )
		
		atfacecentres = Marionette.PortIn(False, 'bFaceCentres')
		atfacecentres.SetDescription( "Place objects at poly face centres" )
	
		classname = Marionette.PortIn('None', 'sClass' )
		classname.SetDescription( "The shape classname used for the EP Extrude object or random classname" )
	
		zoffset = Marionette.PortIn(0, 'nZoffset')
		zoffset.SetDescription( "Z offset of placed objects" )
	
		vertical = Marionette.PortIn(False, 'bVertical')
		vertical.SetDescription( "Placed objects always vertical" )
		
		flipdirection = Marionette.PortIn(False, 'bFlipDir')
		flipdirection.SetDescription( "Placed objects flipped about XY layer plane" )
	
		#Output Ports
		symbols = Marionette.PortOut('hSymbol(s)')
		symbols.SetDescription( "The resultant placed symbols(s)" )
	
		status = Marionette.PortOut('sStatus')
		status.SetDescription( "The status of the created placed object(s)" )
	
	#BEHAVIOR
		this.SetLinksObjects()
	
	def RunNode(self):
	
		def placesymbols(sym,zoffset,classname,atvertices,atmidedges,atfacecentres,vertical,flipdirection,faces,vertices,edges,edgevecs,edgefaces,vertfaces):
	
			def createsym(sym,classname,locdef,zoffset):
				nonlocal symbolpios
				
				axisX,axisY,axisZ,origin = locdef
				if zoffset != 0:
					origin = (origin[0]+axisZ[0]*zoffset,origin[1]+axisZ[1]*zoffset,origin[2]+axisZ[2]*zoffset)
	
				vs.Symbol(sym,(0,0),0)
				symH = vs.LNewObj()
				if vs.SetEntityMatrixN(symH, axisX, axisY, axisZ, origin):
					vs.SetClass(symH,classname) 
					symbolpios.append(symH)
				else:
					vs.DelObject(symH)
				return
			
			def addvec(dir,axis):
				return (dir[0] + axis[0],dir[1] + axis[1],dir[2] + axis[2])
			
			def flipvec(vec):
				return (-vec[0],-vec[1],-vec[2])
	
			# ===== placesymbols
			if atfacecentres:
				for face in faces:
					if vertical:
						xdir = (1,0,0)
						ydir = (0,1,0)
						zdir = (0,0,-1) if flipdirection else (0,0,1)
						origin = face[3]
					else:
						xdir,ydir,zdir,origin = face
	
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
					createsym(sym,classname,(xdir,ydir,zdir,origin),zoffset)
			
			if atvertices:
				if vertical:
					# reset X, Y and Z axes if used
					xdir = (1,0,0); ydir = (0,1,0)
					zdir = (0,0,-1) if flipdirection else (0,0,1)
	
				for vno in range(len(vertices)):
					vertexpos = vertices[vno]
	
					if not vertical:
						xdir = (0,0,0)
						ydir = (0,0,0)
						zdir = (0,0,0)
	
						facenos = vertfaces[vno]
						for j in range(len(facenos)):
							faceno = facenos[j]
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = addvec(zdir,zaxis)
	
						zdir = vs.UnitVec(zdir)
						if vs.EqPt3D((zdir[0],zdir[1],abs(zdir[2])),(0,0,1),0.001): # zdir is vertical
							xdir = (1,0,0) if zdir[2] > 0 else (-1,0,0)
						else:
							xdir = vs.Perp(zdir)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
					createsym(sym,classname,(xdir, ydir, zdir, vertexpos),zoffset)
				
			if atmidedges:
				for i in range(len(edgefaces)):
					edgevec = edgevecs[i]
					midpos = edgevec[0]
					
					if vertical:
						# reset X, Y and Z axes
						xdir = (1,0,0); ydir = (0,1,0)
						zdir = (0,0,-1) if flipdirection else (0,0,1)
					else:
						facedef = edgefaces[i]
						xdir	= edgevec[1]
						zdir = (0,0,0)
						for faceno in facedef:
							xaxis,yaxis,zaxis,centpos = faces[faceno]
							zdir = addvec(zdir,zaxis)
	
						zdir = vs.UnitVec(zdir)
						ydir = vs.CrossProduct(xdir,zdir) # always returns unit vector?
						if not flipdirection:
							ydir = flipvec(ydir)
							zdir = flipvec(zdir)
	
					createsym(sym,classname,(xdir, ydir, zdir, midpos),zoffset)
			return
	
		# ===== RunNode
		status = 0
		symbolpios = []
	
		# Node controls
		symbols  = self.Params.getinputSymboldefs(self.Params.symboldefs.value)
		polydefs = self.Params.getinputPolydefs(self.Params.polydefs.value)
	
		# UI controls
		atvertices = self.Params.getboolval(self.Params.atvertices.value)
		atmidedges = self.Params.getboolval(self.Params.atmidedges.value)
		atfacecentres = self.Params.getboolval(self.Params.atfacecentres.value)
		atset = (atvertices or atmidedges or atfacecentres)
	
		if (symbols and polydefs and atset):
			classname	  = self.Params.getclass(self.Params.classname.value)
			zoffset		  = self.Params.getnumbval(self.Params.zoffset.value,0)
			vertical	  = self.Params.getboolval(self.Params.vertical.value)
			flipdirection = self.Params.getboolval(self.Params.flipdirection.value)
			
			faces	  = polydefs['faces']
			vertices  = polydefs['vertices']
			edges	  = polydefs['edges']
			edgevecs  = polydefs['edgevecs']
			edgefaces = polydefs['edgefaces']
			vertfaces = polydefs['vertfaces']
	
			for sym in symbols:
				placesymbols(sym,zoffset,classname,atvertices,atmidedges,atfacecentres,vertical,flipdirection,faces,vertices,edges,edgevecs,edgefaces,vertfaces)
	
		else:
			if not polydefs:
				status = 'Invalid poly definition(s) or Multi poly definition'
			elif not symbols:
				status = 'Invalid symbol definition(s)'
			elif not atset:
				status = 'No place at settings'
	
		#outputs
		self.Params.symbols.value = symbolpios
		self.Params.status.value = status
	